# Generated by Django 4.2.25 on 2025-10-17 05:40

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "taxsystem",
            "0008_journalfilter_journalfilterset_delete_filteramount_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="payments",
            name="corporation_id",
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.db import models

# AA TaxSystem
from taxsystem.providers import esi

if TYPE_CHECKING:
    # AA TaxSystem
    from taxsystem.models.general import EveEntity


class EveEntityNameResolver:
    """
    Container with a mapping between entity Ids and entity names and a performant API
    """

    def __init__(self, names_map: dict[int, str]) -> None:
        self._names_map = names_map

    def to_name(self, eve_id: int) -> str:
        """Resolved an entity ID to a name

        Args:
            eve_id: ID of the Eve entity to resolve

        Returns:
            name for corresponding entity ID if known else an empty string
        """
        try:
            name = self._names_map[eve_id]
        except KeyError:
            name = ""

        return name


class EveEntityManager(models.Manager["EveEntity"]):
    def bulk_resolve_names(self, ids: list[int]) -> EveEntityNameResolver:
        """Bulk resolve Eve IDs to names and categories using ESI."""
        if not ids:
            return EveEntityNameResolver({})

        names_map = {
            entity.id: entity.name
            for entity in self.filter(id__in=ids).only("id", "name")
        }
        new_ids = set(ids).difference(names_map.keys())

        if new_ids:
            response = esi.client.Universe.PostUniverseNames(
                body=list(new_ids)
            ).results()
            new_entities = []

            for entity_data in response:
                new_entities.append(
                    self.model(
                        id=entity_data.id,
                        name=entity_data.name,
                        category=entity_data.category,
                    )
                )
                names_map[entity_data.id] = entity_data.name

            if new_entities:
                self.bulk_create(new_entities, batch_size=500, ignore_conflicts=True)
        return EveEntityNameResolver(names_map)



# Generated by Django 4.2.26 on 2025-11-14 13:21

# Django
import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("taxsystem", "0009_payments_corporation_id"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="OwnerAudit",
            new_name="CorporationOwner",
        ),
        migrations.RenameModel(
            old_name="Payments",
            new_name="CorporationPayments",
        ),
        migrations.RenameModel(
            old_name="PaymentSystem",
            new_name="CorporationPaymentAccount",
        ),
        migrations.RenameModel(
            old_name="OwnerUpdateStatus",
            new_name="CorporationUpdateStatus",
        ),
        migrations.CreateModel(
            name="AllianceOwner",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("active", models.BooleanField(default=True)),
                (
                    "tax_amount",
                    models.DecimalField(
                        decimal_places=0,
                        default=0,
                        help_text="Tax Amount in ISK that is set for the corporation. Max 16 Digits",
                        max_digits=16,
                        validators=[
                            django.core.validators.MaxValueValidator(9999999999999999)
                        ],
                    ),
                ),
                (
                    "tax_period",
                    models.PositiveIntegerField(
                        default=30,
                        help_text="Tax Period in days for the corporation. Max 365 days. Default: 30 days",
                        validators=[django.core.validators.MaxValueValidator(365)],
                    ),
                ),
                (
                    "eve_alliance",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        help_text="The corporation that owns this alliance tax system.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="taxsystem.corporationowner",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="AllianceOwnerUpdateStatus",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_success",
                    models.BooleanField(db_index=True, default=None, null=True),
                ),
                ("error_message", models.TextField()),
                ("has_token_error", models.BooleanField(default=False)),
                (
                    "last_run_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last run has been started at this time",
                        null=True,
                    ),
                ),
                (
                    "last_run_finished_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last run has been successful finished at this time",
                        null=True,
                    ),
                ),
                (
                    "last_update_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last update has been started at this time",
                        null=True,
                    ),
                ),
                (
                    "last_update_finished_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last update has been successful finished at this time",
                        null=True,
                    ),
                ),
                (
                    "section",
                    models.CharField(
                        choices=[
                            ("payments", "Payments"),
                            ("payment_system", "Payment System"),
                            ("payday", "Payday"),
                        ],
                        db_index=True,
                        max_length=32,
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ts_alliance_update_status",
                        to="taxsystem.allianceowner",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="AlliancePaymentAccount",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("date", models.DateTimeField(auto_now_add=True, null=True)),
                (
                    "status",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("active", "Active"),
                            ("inactive", "Inactive"),
                            ("deactivated", "Deactivated"),
                            ("missing", "Missing"),
                        ],
                        default="active",
                        max_length=16,
                    ),
                ),
                (
                    "deposit",
                    models.DecimalField(
                        decimal_places=0,
                        default=0,
                        help_text="Deposit Pool in ISK. Max 16 Digits",
                        max_digits=16,
                        validators=[
                            django.core.validators.MaxValueValidator(9999999999999999),
                            django.core.validators.MinValueValidator(-9999999999999999),
                        ],
                    ),
                ),
                ("last_paid", models.DateTimeField(blank=True, null=True)),
                ("notice", models.TextField(blank=True, null=True)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ts_alliance_payment_accounts",
                        to="taxsystem.allianceowner",
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="AlliancePayments",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("entry_id", models.BigIntegerField()),
                ("amount", models.DecimalField(decimal_places=0, max_digits=12)),
                ("date", models.DateTimeField(blank=True, null=True)),
                ("reason", models.TextField(blank=True, null=True)),
                (
                    "request_status",
                    models.CharField(
                        choices=[
                            ("approved", "Approved"),
                            ("pending", "Pending"),
                            ("rejected", "Rejected"),
                            ("needs_approval", "Requires Auditor"),
                        ],
                        default="pending",
                        max_length=16,
                        verbose_name="Request Status",
                    ),
                ),
                (
                    "reviser",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Reviser that approved or rejected the payment",
                        max_length=100,
                    ),
                ),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ts_alliance_payments",
                        to="taxsystem.alliancepaymentaccount",
                    ),
                ),
                (
                    "owner_id",
                    models.PositiveIntegerField(
                        help_text=(
                            "ID of the owner (corporation or alliance) associated with this payment"
                        ),
                        null=True,
                        blank=True,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.db import models, transaction
from django.utils import timezone

# Alliance Auth
from allianceauth.authentication.models import User, UserProfile
from allianceauth.services.hooks import get_extension_logger

# AA TaxSystem
from taxsystem import __title__
from taxsystem.app_settings import TAXSYSTEM_BULK_BATCH_SIZE
from taxsystem.decorators import log_timing
from taxsystem.models.helpers.textchoices import (
    AccountStatus,
    AllianceUpdateSection,
    PaymentActions,
    PaymentRequestStatus,
    PaymentSystemText,
)
from taxsystem.providers import AppLogger

logger = AppLogger(get_extension_logger(__name__), __title__)

if TYPE_CHECKING:
    # AA TaxSystem
    from taxsystem.models.alliance import AllianceOwner as OwnerContext
    from taxsystem.models.alliance import (
        AlliancePaymentAccount as PaymentAccountContext,
    )
    from taxsystem.models.alliance import AlliancePayments as PaymentsContext


# TODO Make a all in one manager for both corp and alliance tax accounts?
# pylint: disable=duplicate-code
class AlliancePaymentAccountManager(models.Manager["PaymentAccountContext"]):
    @log_timing(logger)
    def update_or_create_tax_accounts(
        self, owner: "OwnerContext", force_refresh: bool = False
    ) -> None:
        """Update or Create Tax Account data."""
        return owner.update_manager.update_section_if_changed(
            section=AllianceUpdateSection.TAX_ACCOUNTS,
            fetch_func=self._update_or_create_objs,
            force_refresh=force_refresh,
        )

    @transaction.atomic()
    # pylint: disable=unused-argument, too-many-locals
    def _update_or_create_objs(
        self, owner: "OwnerContext", force_refresh: bool = False, runs: int = 0
    ) -> None:
        """Update or Create tax accounts entries from objs data."""
        # pylint: disable=import-outside-toplevel, cyclic-import
        # AA TaxSystem
        from taxsystem.models.alliance import (
            AllianceFilterSet,
            AlliancePaymentHistory,
            AlliancePayments,
        )

        # TODO Create a Hash Tag to track changes better
        logger.debug(
            "Updating Tax Accounts for: %s",
            owner.name,
        )

        payments = AlliancePayments.objects.filter(
            account__owner=owner,
            request_status__in=[
                PaymentRequestStatus.PENDING,
                PaymentRequestStatus.NEEDS_APPROVAL,
            ],
        )

        _current_payment_ids = set(payments.values_list("id", flat=True))
        _automatic_payment_ids = []

        # Check tax accounts before we process payments
        self._check_tax_accounts(owner)

        filters_obj = AllianceFilterSet.objects.filter(owner=owner)
        for filter_obj in filters_obj:
            f_payments = filter_obj.filter(payments)
            for payment in f_payments:
                # Process only pending or needs approval payments
                if payment.request_status in [
                    PaymentRequestStatus.PENDING,
                    PaymentRequestStatus.NEEDS_APPROVAL,
                ]:
                    # Ensure all transfers are processed in a single transaction
                    with transaction.atomic():
                        payment.request_status = PaymentRequestStatus.APPROVED
                        payment.reviser = "System"

                        # Update payment pool for user
                        self.filter(owner=owner, user=payment.account.user).update(
                            deposit=payment.account.deposit + payment.amount
                        )

                        payment.save()

                        AlliancePaymentHistory(
                            user=payment.account.user,
                            payment=payment,
                            action=PaymentActions.STATUS_CHANGE,
                            new_status=PaymentRequestStatus.APPROVED,
                            comment=PaymentSystemText.AUTOMATIC,
                        ).save()

                        runs = runs + 1
                        _automatic_payment_ids.append(payment.pk)

        # Check for any payments that need approval
        needs_approval = _current_payment_ids - set(_automatic_payment_ids)
        approvals = AlliancePayments.objects.filter(
            id__in=needs_approval,
            request_status=PaymentRequestStatus.PENDING,
        )

        for payment in approvals:
            payment.request_status = PaymentRequestStatus.NEEDS_APPROVAL
            payment.save()

            AlliancePaymentHistory(
                user=payment.account.user,
                payment=payment,
                action=PaymentActions.STATUS_CHANGE,
                new_status=PaymentRequestStatus.NEEDS_APPROVAL,
                comment=PaymentSystemText.REVISER,
            ).save()

            runs = runs + 1

        logger.debug(
            "Finished %s: Tax Accounts entrys for %s",
            runs,
            owner.name,
        )

        return ("Finished Tax Accounts for %s", owner.name)

    # pylint: disable=duplicate-code
    def _check_tax_accounts(self, owner: "OwnerContext"):
        """
        Check tax accounts for a alliance.
        Create new accounts, update existing ones, and remove orphaned accounts.
        """
        logger.debug("Checking Tax Accounts for: %s", owner.name)
        items = []

        # Get all existing accounts with a Main Character
        auth_accounts = UserProfile.objects.filter(
            main_character__isnull=False,
        ).prefetch_related("user__profile__main_character")
        auth_accounts_ids = set(auth_accounts.values_list("user_id", flat=True))

        # If no valid accounts, return
        if not auth_accounts:
            logger.debug("No valid accounts for skipping Check: %s", owner.name)
            return "No Accounts"

        # Get existing and new accounts
        existing_accounts = self.filter(owner=owner).select_related("user")
        existing_accounts_ids = set(existing_accounts.values_list("user_id", flat=True))

        # Filter only new accounts
        new_accounts = auth_accounts.exclude(
            user__in=existing_accounts.values_list("user", flat=True)
        )

        # Cleanup orphaned accounts
        self._cleanup_orphaned_accounts(owner, auth_accounts_ids, existing_accounts_ids)

        # Update existing accounts
        for tax_account in existing_accounts:
            self._update_existing_account(tax_account)

        # Create new accounts for users without existing tax accounts
        for account in new_accounts:
            logger.debug(
                "Creating new alliance tax account for user: %s", account.user.username
            )
            items.append(
                self.model(
                    name=account.main_character.character_name,
                    owner=owner,
                    user=account.user,
                    status=AccountStatus.ACTIVE,
                )
            )

        # Bulk create new accounts
        if items:
            self.bulk_create(
                items,
                batch_size=TAXSYSTEM_BULK_BATCH_SIZE,
                ignore_conflicts=True,
            )
            logger.info("Added %s new tax accounts for: %s", len(items), owner.name)
        else:
            logger.debug("No new tax accounts for: %s", owner.name)

        return ("Finished checking Tax Accounts for %s", owner.name)

    # pylint: disable=duplicate-code
    def _cleanup_orphaned_accounts(
        self, owner: "OwnerContext", auth_user_ids: set, ps_user_ids: set
    ):
        """Delete Tax accounts for users without main characters."""
        for ps_user_id in ps_user_ids:
            if ps_user_id not in auth_user_ids:
                self.filter(owner=owner, user_id=ps_user_id).delete()
                logger.info(
                    "Deleted Tax Account for user id: %s from Alliance: %s",
                    ps_user_id,
                    owner.name,
                )

    # pylint: disable=duplicate-code
    def _update_existing_account(
        self,
        tax_account: "PaymentAccountContext",
    ):
        """
        Update an existing tax account based on current state.

        Args:
            owner (AllianceOwner): The owner of the tax account.
            tax_account (AlliancePaymentAccount): The tax account to update.
        """
        # pylint: disable=import-outside-toplevel, cyclic-import
        # AA TaxSystem
        from taxsystem.models.alliance import AllianceOwner

        # Get alliance IDs
        pa_ally_id = tax_account.owner.eve_alliance.alliance_id
        main_ally_id = tax_account.user.profile.main_character.alliance_id

        # Reactivate Account if user returned to alliance
        if tax_account.status == AccountStatus.MISSING and main_ally_id == pa_ally_id:
            self._reset_account(tax_account)
            return

        # Update Account when user left the alliance
        if pa_ally_id != main_ally_id:
            # Mark as missing if not already
            if not tax_account.is_missing:
                self._mark_missing_tax_account(tax_account)
            # Try to move to new alliance if exists
            try:
                new_owner = AllianceOwner.objects.get(
                    eve_alliance__alliance_id=main_ally_id
                )
                # Move to new owner
                self._move_tax_account_to_owner(tax_account, new_owner)
            except AllianceOwner.DoesNotExist:
                pass
            # Save changes
            tax_account.save()

    # pylint: disable=duplicate-code
    def _reset_account(self, tax_account: "PaymentAccountContext"):
        """
        Reset tax account state (unsaved).

        Args:
            tax_account (AlliancePaymentAccount): The tax account to reset.
        """
        tax_account.status = AccountStatus.ACTIVE
        tax_account.notice = None
        tax_account.deposit = 0
        tax_account.last_paid = None
        tax_account.save()
        logger.info(
            "Reset Tax Account %s",
            tax_account.name,
        )

    # pylint: disable=duplicate-code
    def _move_tax_account_to_owner(
        self, tax_account: "PaymentAccountContext", owner: "OwnerContext"
    ):
        """
        Move a tax account to another owner.
        Resets account state.

        Args:
            tax_account (AlliancePaymentAccount): The tax account to move.
            owner (AllianceOwner): The new owner of the tax account.
        """
        tax_account.owner = owner
        tax_account.status = AccountStatus.ACTIVE
        tax_account.notice = None
        tax_account.deposit = 0
        tax_account.last_paid = None
        tax_account.save()
        logger.info(
            "Moved Tax Account %s to Alliance %s",
            tax_account.name,
            owner.eve_alliance.alliance_name,
        )

    # pylint: disable=duplicate-code
    def _mark_missing_tax_account(self, tax_account: "PaymentAccountContext"):
        """
        Mark the tax account as missing (unsaved).

        Args:
            tax_account (AlliancePaymentAccount): The tax account to mark as missing.
        """
        tax_account.status = AccountStatus.MISSING
        tax_account.save()
        logger.info("Marked Tax Account %s as MISSING", tax_account.name)

    @log_timing(logger)
    def check_payment_deadlines(
        self, owner: "OwnerContext", force_refresh: bool = False
    ) -> None:
        """Check Payments from Account."""
        return owner.update_manager.update_section_if_changed(
            section=AllianceUpdateSection.DEADLINES,
            fetch_func=self._payment_deadlines,
            force_refresh=force_refresh,
        )

    @transaction.atomic()
    # pylint: disable=unused-argument
    def _payment_deadlines(
        self, owner: "OwnerContext", force_refresh: bool = False
    ) -> None:
        """
        Checking payment deadlines for Alliance.
        This will deduct tax amounts from deposits if payment period has passed.
        """
        logger.debug(
            "Updating payment deadlines for: %s",
            owner.name,
        )

        tax_accounts = self.filter(owner=owner, status=AccountStatus.ACTIVE)

        items = []
        for account in tax_accounts:
            if account.last_paid is None:
                # First Period is free
                account.last_paid = timezone.now()
            if timezone.now() - account.last_paid >= timezone.timedelta(
                days=owner.tax_period
            ):
                account.deposit -= owner.tax_amount
                account.last_paid = timezone.now()
            items.append(account)

        if not items:
            logger.debug("No new payment deadlines for: %s", owner.name)
            return ("No new payment deadlines for %s", owner.name)

        self.bulk_update(
            items,
            ["deposit", "last_paid"],
            batch_size=TAXSYSTEM_BULK_BATCH_SIZE,
        )

        logger.debug(
            "Finished %s: payment deadlines for %s",
            len(items),
            owner.name,
        )

        return ("Finished payment deadlines for %s", owner.name)


class AlliancePaymentsQuerySet(models.QuerySet["PaymentsContext"]):
    # pylint: disable=duplicate-code
    def visible_to(self, user: User):
        """Return visible payments for the user depending on their permissions."""
        # superusers get all visible
        if user.is_superuser:
            logger.debug(
                "Returning all payments for superuser %s.",
                user,
            )
            return self

        if user.has_perm("taxsystem.manage_alliances"):
            logger.debug(
                "Returning all alliance payments for Tax Audit Manager %s.", user
            )
            return self
        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(account__user=user)]

            if user.has_perm("taxsystem.manage_own_alliance"):
                alliance_ids = user.character_ownerships.all().values_list(
                    "character__alliance_id", flat=True
                )
                queries.append(
                    models.Q(owner__eve_alliance__alliance_id__in=alliance_ids)
                )
            logger.debug("%s queries for user %s visible alliance.", len(queries), user)

            if len(queries) == 0:
                return self.none()

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()

    def visible_open_invoices(self, user: User) -> int | None:
        """
        Get the count of visible invoices for the given user.
        """
        owners = self.visible_to(user=user)
        return owners.filter(
            request_status__in=[
                PaymentRequestStatus.PENDING,
                PaymentRequestStatus.NEEDS_APPROVAL,
            ],
        ).count()

    def open_invoices(self, owner: "OwnerContext") -> int | None:
        """
        Get open invoices for the given owner.
        """
        return self.filter(
            owner=owner,
            request_status__in=[
                PaymentRequestStatus.PENDING,
                PaymentRequestStatus.NEEDS_APPROVAL,
            ],
        ).count()


class AlliancePaymentManager(models.Manager["PaymentsContext"]):
    def get_queryset(self):
        return AlliancePaymentsQuerySet(self.model, using=self._db)

    def get_visible(self, user: User):
        return self.get_queryset().visible_to(user=user)

    def get_visible_open_invoices(self, user: User):
        """Get the count of visible open invoices for the given user."""
        return self.get_queryset().visible_open_invoices(user=user)

    def get_owner_open_invoices(self, user: User, owner: "OwnerContext"):
        """Get the count of open invoices for the given user and owner."""
        if user.has_perm("taxsystem.manage_own_alliance") or user.has_perm(
            "taxsystem.manage_alliances"
        ):
            return self.get_queryset().open_invoices(owner=owner)
        return 0

    @log_timing(logger)
    def update_or_create_payments(
        self, owner: "OwnerContext", force_refresh: bool = False
    ) -> None:
        """Update or Create a Payments entry data."""
        return owner.update_manager.update_section_if_changed(
            section=AllianceUpdateSection.PAYMENTS,
            fetch_func=self._update_or_create_objs,
            force_refresh=force_refresh,
        )

    @transaction.atomic()
    # pylint: disable=too-many-locals, unused-argument, duplicate-code
    def _update_or_create_objs(
        self, owner: "OwnerContext", force_refresh: bool = False
    ) -> None:
        """Update or Create payments for Alliance."""
        # pylint: disable=import-outside-toplevel, cyclic-import
        # AA TaxSystem
        from taxsystem.models.alliance import AlliancePaymentAccount as PaymentAccount
        from taxsystem.models.alliance import AlliancePaymentHistory, AlliancePayments
        from taxsystem.models.wallet import CorporationWalletJournalEntry

        logger.debug(
            "Updating payments for: %s",
            owner.name,
        )

        tax_accounts = PaymentAccount.objects.filter(owner=owner)

        if not tax_accounts:
            return ("No Payment Users for %s", owner.name)

        users = {}

        # Build Account Map
        for account in tax_accounts:
            account: PaymentAccount
            alts = account.get_alt_ids()
            users[account] = alts

        # Check journal entries for player donations
        journal_qs = CorporationWalletJournalEntry.objects.filter(
            division__corporation=owner.corporation,
            ref_type__in=["player_donation"],
        ).order_by("-date")

        _current_entry_ids = set(
            self.filter(account__owner=owner).values_list(
                "journal__entry_id", flat=True
            )
        )

        with transaction.atomic():
            items = []
            logs_items = []
            for journal in journal_qs:
                # Skip if already processed
                if journal.entry_id in _current_entry_ids:
                    continue

                # Ensure first party is not null to avoid issues with journal entries without a first party
                try:
                    assert journal.first_party
                except AssertionError:
                    logger.warning(
                        "Journal entry %s has no first party. Skipping.",
                        journal.entry_id,
                    )
                    continue

                for account, alts in users.items():
                    # Check if entry belongs to user's characters
                    if journal.first_party.id in alts:
                        payment_item = AlliancePayments(
                            owner=owner,
                            journal=journal,
                            name=account.name,
                            account=account,
                            amount=journal.amount,
                            request_status=PaymentRequestStatus.PENDING,
                            date=journal.date,
                            reason=journal.reason,
                        )
                        items.append(payment_item)

            if not items:
                logger.debug("No new Payments for: %s", owner.name)
                return ("No new Payments for %s", owner.name)

            # Bulk create payments
            new_payments = self.bulk_create(
                items, batch_size=TAXSYSTEM_BULK_BATCH_SIZE, ignore_conflicts=True
            )

            for payment in new_payments:
                # Only log created payments
                payment_obj = self.filter(
                    journal=payment.journal,
                    account=payment.account,
                    owner=owner,
                ).first()

                if not payment_obj:
                    continue

                # Use the saved payment object when creating history entries
                log_items = AlliancePaymentHistory(
                    user=payment.account.user,
                    payment=payment_obj,
                    action=PaymentActions.STATUS_CHANGE,
                    new_status=PaymentRequestStatus.PENDING,
                    comment=PaymentSystemText.ADDED,
                )
                logs_items.append(log_items)

            # Bulk create log entries once after iterating new payments
            if logs_items:
                AlliancePaymentHistory.objects.bulk_create(
                    logs_items,
                    batch_size=TAXSYSTEM_BULK_BATCH_SIZE,
                    ignore_conflicts=True,
                )

        logger.debug(
            "Finished %s Payments for %s",
            len(items),
            owner.name,
        )
        return (
            "Finished %s Payments for %s",
            len(items),
            owner.name,
        )



"""Models for Tax System."""

# Django
from django.core.validators import MaxValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import (
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenError
from esi.models import Token

# AA TaxSystem
from taxsystem import __title__
from taxsystem.managers.corporation_manager import (
    CorporationAccountManager,
    MembersManager,
    PaymentsManager,
)
from taxsystem.managers.owner_manager import (
    CorporationOwnerManager,
)
from taxsystem.models.base import (
    FilterBaseModel,
    FilterSetBaseModel,
    HistoryBaseModel,
    PaymentAccountBaseModel,
    PaymentsBaseModel,
    UpdateStatusBaseModel,
)
from taxsystem.models.general import (
    UpdateSectionResult,
)
from taxsystem.models.helpers.textchoices import (
    AdminActions,
    CorporationUpdateSection,
    PaymentRequestStatus,
    UpdateStatus,
)
from taxsystem.models.helpers.updater import UpdateManager
from taxsystem.models.wallet import (
    CorporationWalletDivision,
    CorporationWalletJournalEntry,
)
from taxsystem.providers import AppLogger, esi

logger = AppLogger(get_extension_logger(__name__), __title__)


class CorporationUpdateStatus(UpdateStatusBaseModel):
    """Model representing the update status of a corporation owner in the tax system."""

    owner = models.ForeignKey(
        "CorporationOwner",
        on_delete=models.CASCADE,
        related_name="ts_corporation_update_status",
    )
    section = models.CharField(
        max_length=32, choices=CorporationUpdateSection.choices, db_index=True
    )

    class Meta:
        default_permissions = ()
        unique_together = [("owner", "section")]

    def __str__(self) -> str:
        return f"{self.owner} - {self.section}"


class CorporationOwner(models.Model):
    """Model representing a corporation owner in the tax system."""

    class Meta:
        default_permissions = ()

    objects: CorporationOwnerManager = CorporationOwnerManager()

    name = models.CharField(
        max_length=255,
    )

    eve_corporation = models.OneToOneField(
        EveCorporationInfo, on_delete=models.CASCADE, related_name="+"
    )

    active = models.BooleanField(
        default=True,
        help_text=_("Designates whether this corporation owner is active."),
    )

    tax_amount = models.DecimalField(
        max_digits=16,
        decimal_places=0,
        help_text=_("Tax Amount in ISK that is set for the corporation. Max 16 Digits"),
        default=0,
        validators=[MaxValueValidator(9999999999999999)],
    )

    tax_period = models.PositiveIntegerField(
        help_text=_(
            "Tax Period in days for the corporation. Max 365 days. Default: 30 days"
        ),
        default=30,
        validators=[MaxValueValidator(365)],
    )

    def __str__(self):
        return f"{self.name}"

    @property
    def payment_model(self):
        """Return the Payment Model for this owner."""
        return CorporationPayments

    @property
    def payment_history_model(self):
        """Return the Payment History Model for this owner."""
        return CorporationPaymentHistory

    @property
    def admin_log_model(self):
        """Return the Admin History Model for this owner."""
        return CorporationAdminHistory

    @property
    def account_model(self):
        """Return the Tax Account Model for this owner."""
        return CorporationPaymentAccount

    @property
    def filterset_model(self):
        """Return the Filter Set Model for this owner."""
        return CorporationFilterSet

    @property
    def filter_model(self):
        """Return the Filter Model for this owner."""
        return CorporationFilter

    @property
    def update_manager(self):
        """Return the Update Manager helper for this owner."""
        return UpdateManager(
            owner=self,
            update_section=CorporationUpdateSection,
            update_status=CorporationUpdateStatus,
        )

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            # General
            "esi-corporations.read_corporation_membership.v1",
            "esi-corporations.track_members.v1",
            "esi-characters.read_corporation_roles.v1",
            # wallets
            "esi-wallet.read_corporation_wallets.v1",
            "esi-corporations.read_divisions.v1",
        ]

    def get_token(self, scopes, req_roles) -> Token:
        """Get the token for this corporation."""
        if "esi-characters.read_corporation_roles.v1" not in scopes:
            scopes.append("esi-characters.read_corporation_roles.v1")

        char_ids = EveCharacter.objects.filter(
            corporation_id=self.eve_corporation.corporation_id
        ).values("character_id")

        tokens = Token.objects.filter(character_id__in=char_ids).require_scopes(scopes)

        for token in tokens:
            try:
                roles = esi.client.Character.GetCharactersCharacterIdRoles(
                    character_id=token.character_id, token=token
                ).result(force_refresh=True)

                has_roles = False
                for role in roles.roles:
                    if role in req_roles:
                        has_roles = True

                if has_roles:
                    return token
            except TokenError as e:
                logger.error(
                    "Token ID: %s (%s)",
                    token.pk,
                    e,
                )
        return False

    def update_division_names(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the divisions for this corporation."""
        return CorporationWalletDivision.objects.update_or_create_esi_names(
            owner=self, force_refresh=force_refresh
        )

    def update_divisions(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the divisions for this corporation."""
        return CorporationWalletDivision.objects.update_or_create_esi(
            owner=self, force_refresh=force_refresh
        )

    def update_wallet(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the wallet journals for this corporation."""
        return CorporationWalletJournalEntry.objects.update_or_create_esi(
            owner=self, force_refresh=force_refresh
        )

    def update_members(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the members for this corporation.
        Args:
            force_refresh: Force refresh from ESI even if not modified
        Returns:
            UpdateSectionResult object for this section
        """
        return Members.objects.update_or_create_esi(self, force_refresh=force_refresh)

    def update_payments(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the payments for this owner.
        Args:
            force_refresh: Force refresh
        Returns:
            UpdateSectionResult object for this section
        """
        return CorporationPayments.objects.update_or_create_payments(
            owner=self, force_refresh=force_refresh
        )

    def update_tax_accounts(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the tax accounts for this owner.
        Args:
            force_refresh: Force refresh
        Returns:
            UpdateSectionResult object for this section
        """
        return CorporationPaymentAccount.objects.update_or_create_tax_accounts(
            owner=self, force_refresh=force_refresh
        )

    def update_deadlines(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the tax deadlines for this owner.
        Args:
            force_refresh: Force refresh
        Returns:
            UpdateSectionResult object for this section
        """
        return CorporationPaymentAccount.objects.check_payment_deadlines(
            owner=self, force_refresh=force_refresh
        )

    # Abstract properties implementation
    @property
    def eve_id(self) -> int:
        """Return the Eve Corporation ID."""
        return self.eve_corporation.corporation_id

    @property
    def get_status(self) -> UpdateStatus:
        """Get the update status of this owner.

        Returns:
            UpdateStatus enum value representing the current status
        """
        if self.active is False:
            return UpdateStatus.DISABLED

        # Use type(self) for dynamic QuerySet resolution
        qs = type(self).objects.filter(pk=self.pk).annotate_total_update_status()
        total_update_status = list(qs.values_list("total_update_status", flat=True))[0]
        return UpdateStatus(total_update_status)

    @property
    def get_update_status(self) -> dict[str, str]:
        """Return a dictionary of update sections and their statuses."""
        update_status = {}
        for section in CorporationUpdateSection.get_sections():
            try:
                status = CorporationUpdateStatus.objects.get(
                    owner=self, section=section
                )
                update_status[section] = {
                    "is_success": status.is_success,
                    "last_update_finished_at": status.last_update_finished_at,
                    "last_run_finished_at": status.last_run_finished_at,
                }
            except CorporationUpdateStatus.DoesNotExist:
                continue
        return update_status


class Members(models.Model):
    """Tax System Member model for app"""

    class Meta:
        default_permissions = ()

    objects: MembersManager = MembersManager()

    class States(models.TextChoices):
        ACTIVE = "active", _("Active")
        MISSING = "missing", _("Missing")
        NOACCOUNT = "noaccount", _("Unregistered")
        IS_ALT = "is_alt", _("is alt")

    character_name = models.CharField(max_length=100, db_index=True)

    character_id = models.PositiveIntegerField(primary_key=True)

    owner = models.ForeignKey(
        CorporationOwner, on_delete=models.CASCADE, related_name="ts_members"
    )

    status = models.CharField(
        _("Status"), max_length=10, choices=States.choices, blank=True, default="active"
    )

    logon = models.DateTimeField(null=True, blank=True)

    logged_off = models.DateTimeField(null=True, blank=True)

    joined = models.DateTimeField(null=True, blank=True)

    notice = models.TextField(null=True, blank=True)

    def __str__(self):
        return f"{self.character_name} - {self.character_id}"

    @property
    def is_active(self) -> bool:
        return self.status == self.States.ACTIVE

    @property
    def is_missing(self) -> bool:
        return self.status == self.States.MISSING

    @property
    def is_noaccount(self) -> bool:
        return self.status == self.States.NOACCOUNT

    @property
    def is_alt(self) -> bool:
        return self.status == self.States.IS_ALT

    @property
    def is_faulty(self) -> bool:
        return self.status in [self.States.MISSING, self.States.NOACCOUNT]


class CorporationPaymentAccount(PaymentAccountBaseModel):
    """Model representing a corporation tax account in the tax system."""

    objects: CorporationAccountManager = CorporationAccountManager()

    class Meta:
        default_permissions = ()

    owner = models.ForeignKey(
        CorporationOwner,
        on_delete=models.CASCADE,
        related_name="ts_corp_tax_accounts",
    )


class CorporationPayments(PaymentsBaseModel):
    """Model representing payments made by corporation members in the tax system."""

    class Meta:
        default_permissions = ()

    objects: PaymentsManager = PaymentsManager()

    account = models.ForeignKey(
        CorporationPaymentAccount,
        on_delete=models.CASCADE,
        related_name="+",
    )

    owner = models.ForeignKey(
        CorporationOwner,
        on_delete=models.CASCADE,
        related_name="ts_corporation_payments",
        help_text=_("Owner of this payment"),
        null=True,
        blank=True,
    )

    def __str__(self) -> str:
        return f"{self.account.name} - {self.amount} ISK"

    # pylint: disable=duplicate-code
    def transaction_log(
        self, user, comment, new_status, action=""
    ) -> "CorporationPaymentHistory":
        """Return a log entry for the transaction.

        Args:
            user: User performing the action
            comment: Additional message
            new_status: New status after the action
            action: Action performed (optional)
        Returns:
            CorporationPaymentHistory object
        """
        return CorporationPaymentHistory(
            user=user,
            payment=self,
            new_status=new_status,
            action=action,
            comment=comment,
        )


class CorporationFilterSet(FilterSetBaseModel):
    class Meta:
        default_permissions = ()

    owner = models.ForeignKey(
        CorporationOwner,
        on_delete=models.CASCADE,
        related_name="ts_corporation_filter_set",
    )

    def filter(
        self, payments: models.QuerySet[CorporationPayments]
    ) -> models.QuerySet[CorporationPayments]:
        """Apply filters to the given payments queryset."""
        if self.is_active:
            queries = []
            for f in self.ts_corporation_filters.all():
                f: CorporationFilter
                # Generate Q object for each filter
                q = f.get_match_type_filter()
                if q is not None:
                    queries.append(q)

            # If no queries were generated, return empty queryset
            if not queries:
                return CorporationPayments.objects.none()

            # Combine all queries using AND operation
            combined = queries.pop()
            for q in queries:
                combined &= q
            return payments.filter(combined)
        return CorporationPayments.objects.none()

    def __str__(self) -> str:
        return f"Filter Set: {self.name}"


class CorporationFilter(FilterBaseModel):
    class Meta:
        default_permissions = ()

    filter_set = models.ForeignKey(
        CorporationFilterSet,
        on_delete=models.CASCADE,
        related_name="ts_corporation_filters",
    )


class CorporationPaymentHistory(HistoryBaseModel):
    """Model representing the history of actions taken on corporation payments in the tax system."""

    class Meta:
        default_permissions = ()

    # pylint: disable=duplicate-code
    payment = models.ForeignKey(
        CorporationPayments,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("Payment"),
        help_text=_("Payment that the action was performed on"),
    )

    # pylint: disable=duplicate-code
    new_status = models.CharField(
        max_length=20,
        choices=PaymentRequestStatus.choices,
        verbose_name=_("New Status"),
        help_text=_("New Status of the action"),
    )


class CorporationAdminHistory(HistoryBaseModel):
    """
    Model representing the history of administrative actions taken on owners in the tax system.
    """

    class Meta:
        default_permissions = ()

    # pylint: disable=duplicate-code
    owner = models.ForeignKey(
        CorporationOwner,
        verbose_name=_("Owner"),
        help_text=_("Owner that the action was performed on"),
        on_delete=models.CASCADE,
        related_name="ts_admin_history",
    )

    # pylint: disable=duplicate-code
    action = models.CharField(
        max_length=20,
        choices=AdminActions.choices,
        default=AdminActions.DEFAULT,
        verbose_name=_("Action"),
        help_text=_("Action performed"),
    )



# Generated by Django 4.2.26 on 2025-11-22 21:54

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taxsystem", "0017_alter_allianceupdatestatus_unique_together_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="allianceowner",
            name="corporation",
            field=models.ForeignKey(
                help_text="The corporation that owns this alliance tax system.",
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="taxsystem.corporationowner",
            ),
        ),
    ]



# Generated by Django 4.2.28 on 2026-02-25 19:32

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taxsystem", "0024_alter_allianceadminhistory_action_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="EveEntity",
            fields=[
                ("id", models.BigIntegerField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=255)),
                ("category", models.CharField(max_length=50)),
                ("last_updated", models.DateTimeField(auto_now=True)),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.AddField(
            model_name="corporationwalletjournalentry",
            name="first_party_new",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="taxsystem.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="corporationwalletjournalentry",
            name="second_party_new",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="taxsystem.eveentity",
            ),
        ),
    ]



# Standard Library
import logging

# Third Party
from ninja import NinjaAPI
from ninja.security import django_auth

# Django
from django.conf import settings

# AA TaxSystem
from taxsystem.api import admin, corporation, filters, logs, payments

api = NinjaAPI(
    title="TaxSystem API",
    version="0.5.0",
    urls_namespace="taxsystem:api",
    auth=django_auth,
    openapi_url=settings.DEBUG and "/openapi.json" or "",
)


def setup(ninja_api):
    corporation.CorporationApiEndpoints(ninja_api)
    admin.AdminApiEndpoints(ninja_api)
    payments.PaymentsApiEndpoints(ninja_api)
    logs.LogsApiEndpoints(ninja_api)
    filters.FilterApiEndpoints(ninja_api)


# Initialize API endpoints
setup(api)



"""Tests for the providers module."""

# Standard Library
from unittest.mock import MagicMock, patch

# Third Party
from aiopenapi3 import RequestError

# Django
from django.test import override_settings

# Alliance Auth
from esi.exceptions import (
    ESIBucketLimitException,
    ESIErrorLimitException,
    HTTPClientError,
    HTTPServerError,
)

# AA TaxSystem
from taxsystem.providers import retry_task_on_esi_error
from taxsystem.tests import NoSocketsTestCase


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestRetryTaskOnESIError(NoSocketsTestCase):
    """Tests for retry_task_on_esi_error context manager."""

    def setUp(self):
        """
        Set up test case with a mock Celery task.
        """
        super().setUp()
        self.task = MagicMock()
        self.task.request.retries = 1
        self.task.retry = MagicMock(side_effect=Exception("Retry called"))

    @patch("taxsystem.providers.random.uniform")
    def test_should_retry_on_esi_error_limit_exception(self, mock_random):
        """
        Test should retry task when ESI error limit is reached.

        Results:
        - The task.retry method is called with appropriate countdown.
        """
        # Test Data
        mock_random.return_value = 3.0  # Fixed jitter for testing
        reset_time = 60.0
        exc = ESIErrorLimitException(reset_time)

        # Test Action
        with self.assertRaises(Exception) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.assertEqual(str(context.exception), "Retry called")
        self.task.retry.assert_called_once()
        call_kwargs = self.task.retry.call_args[1]
        self.assertEqual(call_kwargs["exc"], exc)
        self.assertEqual(call_kwargs["countdown"], 63)

    @patch("taxsystem.providers.random.uniform")
    def test_should_retry_on_esi_bucket_limit_exception(self, mock_random):
        """
        Test should retry task when ESI bucket limit is reached.

        Results:
        - The task.retry method is called with appropriate countdown.
        """
        # Test Data
        mock_random.return_value = 4.0  # Fixed jitter for testing
        reset_time = 30.0
        bucket_name = "test_bucket"
        exc = ESIBucketLimitException(bucket=bucket_name, reset=reset_time)

        # Test Action
        with self.assertRaises(Exception) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.assertEqual(str(context.exception), "Retry called")
        self.task.retry.assert_called_once()
        call_kwargs = self.task.retry.call_args[1]
        self.assertEqual(call_kwargs["exc"], exc)
        self.assertEqual(call_kwargs["countdown"], 34)

    @patch("taxsystem.providers.random.uniform")
    def test_should_retry_on_http_502_error(self, mock_random):
        """
        Test should retry task on HTTP 502 Bad Gateway error.

        Results:
        - The task.retry method is called with appropriate countdown.
        """
        # Test Data
        mock_random.return_value = 2.5  # Fixed jitter for testing
        exc = HTTPServerError(502, {}, b"Bad Gateway")

        # Test Action
        with self.assertRaises(Exception) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.assertEqual(str(context.exception), "Retry called")
        self.task.retry.assert_called_once()
        call_kwargs = self.task.retry.call_args[1]
        self.assertEqual(call_kwargs["exc"], exc)
        self.assertEqual(call_kwargs["countdown"], 62)

    @patch("taxsystem.providers.random.uniform")
    def test_should_retry_on_http_503_error(self, mock_random):
        """
        Test should retry task on HTTP 503 Service Unavailable error.

        Results:
        - The task.retry method is called with appropriate countdown.
        """
        # Test Data
        mock_random.return_value = 3.5  # Fixed jitter for testing
        exc = HTTPServerError(503, {}, b"Service Unavailable")

        # Test Action
        with self.assertRaises(Exception) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.assertEqual(str(context.exception), "Retry called")
        self.task.retry.assert_called_once()
        call_kwargs = self.task.retry.call_args[1]
        self.assertEqual(call_kwargs["countdown"], 63)

    @patch("taxsystem.providers.random.uniform")
    def test_should_retry_on_http_504_error(self, mock_random):
        """
        Test should retry task on HTTP 504 Gateway Timeout error.

        Results:
        - The task.retry method is called with appropriate countdown.
        """
        # Test Data
        mock_random.return_value = 2.0  # Fixed jitter for testing
        exc = HTTPServerError(504, {}, b"Gateway Timeout")

        # Test Action
        with self.assertRaises(Exception) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.assertEqual(str(context.exception), "Retry called")
        self.task.retry.assert_called_once()
        call_kwargs = self.task.retry.call_args[1]
        self.assertEqual(call_kwargs["countdown"], 62)

    @patch("taxsystem.providers.random.uniform")
    def test_should_retry_on_request_error(self, mock_random):
        """
        Test should retry task on Request Error.

        Results:
        - The task.retry method is called with appropriate countdown.
        """
        # Test Data
        mock_random.return_value = 2.0  # Fixed jitter for testing
        exc = RequestError(
            operation=None,
            request=None,
            data="Test Data",
            parameters={},
        )

        # Test Action
        with self.assertRaises(Exception) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.assertEqual(str(context.exception), "Retry called")
        self.task.retry.assert_called_once()
        call_kwargs = self.task.retry.call_args[1]
        self.assertEqual(call_kwargs["countdown"], 62)

    def test_should_not_retry_on_http_404_error(self):
        """
        Test should not retry task on HTTP 404 error (client error).

        Results:
        - The original HTTPClientError is raised.
        - The task.retry method is not called.
        """
        # Test Data
        exc = HTTPClientError(404, {}, b"Not Found")

        # Test Action
        with self.assertRaises(HTTPClientError) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.task.retry.assert_not_called()
        self.assertEqual(context.exception.status_code, 404)

    def test_should_not_retry_on_http_400_error(self):
        """
        Test should not retry task on HTTP 400 error (client error).

        Results:
        - The original HTTPClientError is raised.
        - The task.retry method is not called.
        """
        # Test Data
        exc = HTTPClientError(400, {}, b"Bad Request")

        # Test Action
        with self.assertRaises(HTTPClientError) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.task.retry.assert_not_called()
        self.assertEqual(context.exception.status_code, 400)

    @patch("taxsystem.providers.random.uniform")
    def test_should_apply_backoff_jitter_on_retries(self, mock_random):
        """
        Test should apply exponential backoff jitter based on retry count.

        Results:
        - The countdown for retry should include jitter based on the number of retries.
        """
        # Test Data
        mock_random.return_value = 4.0
        self.task.request.retries = 2  # Third attempt
        reset_time = 60.0
        exc = ESIErrorLimitException(reset_time)

        # Test Action
        with self.assertRaises(Exception):
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        call_kwargs = self.task.retry.call_args[1]
        self.assertEqual(call_kwargs["countdown"], 76)

    def test_should_pass_through_on_success(self):
        """
        Test should pass through when no exception is raised.

        Results:
        - The code inside the context manager executes successfully.
        - The task.retry method is not called.
        """
        # Test Action
        with retry_task_on_esi_error(self.task):
            result = "success"

        # Expected Result
        self.assertEqual(result, "success")
        self.task.retry.assert_not_called()

    def test_should_pass_through_unhandled_exceptions(self):
        """
        Test should pass through exceptions that are not ESI-related.

        Results:
        - The original exception is raised.
        - The task.retry method is not called.
        """
        # Test Data
        exc = ValueError("Some other error")

        # Test Action
        with self.assertRaises(ValueError) as context:
            with retry_task_on_esi_error(self.task):
                raise exc

        # Expected Result
        self.task.retry.assert_not_called()
        self.assertEqual(str(context.exception), "Some other error")



# Standard Library
from unittest.mock import patch

# Django
from django.test import override_settings

# AA TaxSystem
from taxsystem.models.general import EveEntity
from taxsystem.models.general import EveEntity as EveEntityV2
from taxsystem.tests import TaxSystemTestCase
from taxsystem.tests.testdata.esi_stub_openapi import (
    EsiEndpoint,
    create_esi_client_stub,
)
from taxsystem.tests.testdata.utils import (
    create_division,
    create_filterset,
    create_owner_from_user,
)

MODULE_PATH = "taxsystem.managers.eveonline_manager"

TAXSYSTEM_UNIVERSE_ENDPOINTS = [
    EsiEndpoint("Universe", "PostUniverseNames", "ids"),
]


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestEveEntityManager(TaxSystemTestCase):
    """Test Eve Entity Managers."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.audit = create_owner_from_user(cls.user, tax_type="alliance")

        cls.filter_set = create_filterset(
            owner=cls.audit,
            name="Test Filter Set",
            description="Filter Set for Testing Alliance Manager",
        )

        cls.division = create_division(
            corporation=cls.audit.corporation,
            division_id=1,
            name="Main Division",
            balance=1000000,
        )

        cls.eve_character_first_party = EveEntity.objects.get(id=1002)
        cls.eve_character_second_party = EveEntity.objects.get(id=1001)

    @patch(MODULE_PATH + ".esi")
    def test_bulk_resolve_names(self, mock_esi):
        """
        Test bulk_resolve_names method of EveEntityManager with mocked ESI response.
        """
        mock_esi.client = create_esi_client_stub(endpoints=TAXSYSTEM_UNIVERSE_ENDPOINTS)
        manager = EveEntityV2.objects
        # Mock ESI response for universe names
        ids_to_resolve = [9997, 9998]  # Include an ID that doesn't exist
        resolver = manager.bulk_resolve_names(ids_to_resolve)

        # Assert that the resolver returns correct names for known IDs
        self.assertEqual(resolver.to_name(9997), "Bulk Character")
        # Assert that the resolver returns an empty string for unknown ID
        self.assertEqual(resolver.to_name(9998), "Bulk Character 2")

    @patch(MODULE_PATH + ".esi")
    def test_bulk_resolve_names_with_existing_ids_only(self, mock_esi):
        """Should resolve names from existing DB entities when no new IDs are needed."""
        manager = EveEntityV2.objects

        resolver = manager.bulk_resolve_names([1001, 1002])

        self.assertEqual(resolver.to_name(1001), self.eve_character_second_party.name)
        self.assertEqual(resolver.to_name(1002), self.eve_character_first_party.name)
        mock_esi.client.Universe.PostUniverseNames.assert_not_called()

    @patch(MODULE_PATH + ".esi")
    def test_bulk_resolve_names_with_mixed_existing_and_new_ids(self, mock_esi):
        """Should resolve both existing and newly fetched entities in one resolver."""
        mock_esi.client = create_esi_client_stub(endpoints=TAXSYSTEM_UNIVERSE_ENDPOINTS)
        manager = EveEntityV2.objects

        resolver = manager.bulk_resolve_names([1001, 9997])

        self.assertEqual(resolver.to_name(1001), self.eve_character_second_party.name)
        self.assertEqual(resolver.to_name(9997), "Bulk Character")



from .alliance import (
    AllianceAdminHistory,
    AllianceFilter,
    AllianceFilterSet,
    AllianceOwner,
    AlliancePaymentAccount,
    AlliancePaymentHistory,
    AlliancePayments,
    AllianceUpdateStatus,
)
from .corporation import (
    CorporationAdminHistory,
    CorporationFilter,
    CorporationFilterSet,
    CorporationOwner,
    CorporationPaymentAccount,
    CorporationPaymentHistory,
    CorporationPayments,
    CorporationUpdateStatus,
    Members,
)
from .general import General
from .wallet import CorporationWalletDivision, CorporationWalletJournalEntry



# Standard Library
import socket

# Django
from django.test import RequestFactory, TestCase

# AA TaxSystem
from taxsystem.tests.testdata.integrations.allianceauth import load_allianceauth
from taxsystem.tests.testdata.utils import create_user_from_evecharacter


class SocketAccessError(Exception):
    """Error raised when a test script accesses the network"""


class NoSocketsTestCase(TestCase):
    """Variation of Django's TestCase class that prevents any network use.

    Example:

        .. code-block:: python

            class TestMyStuff(BaseTestCase):
                def test_should_do_what_i_need(self): ...

    """

    @classmethod
    def setUpClass(cls):
        cls.socket_original = socket.socket
        socket.socket = cls.guard
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        socket.socket = cls.socket_original
        return super().tearDownClass()

    @staticmethod
    def guard(*args, **kwargs):
        raise SocketAccessError("Attempted to access network")


class TaxSystemTestCase(NoSocketsTestCase):
    """
    Preloaded Testcase class for TaxSystem tests without Network access.

    Pre-Load:
        * Alliance Auth Characters, Corporation, Alliance Data
        * Taken User IDs: 1001, 1002, 1003, 1004, 1005

    Available Request Factory:
        `self.factory`

    Available test users:
        * `user` User with standard TaxSystem access.
            * 'taxsystem.basic_access' Permission
            * Character ID 1001
            * Corporation ID 2001
        * `user2` Second user with standard TaxSystem access.
            * 'taxsystem.basic_access' Permission
            * Character ID 1002
            * Corporation ID 2002
        * `superuser` Superuser.
            * Access to whole Application
            * Character ID 1003
        * `manage_own_user` User with manage own corporation access.
            * 'taxsystem.basic_access' Permission
            * 'taxsystem.manage_own_corp' Permission
            * 'taxsystem.manage_own_alliances' Permission
            * Character ID 1004
            * Corporation ID 2001
            * Alliance ID 3001
        * `manage_user` User with manage corporations access.
            * 'taxsystem.basic_access' Permission
            * 'taxsystem.manage_corps' Permission
            * 'taxsystem.manage_alliances' Permission
            * Character ID 1005
            * Corporation ID 2001
            * Alliance ID 3001

    Example:
        .. code-block:: python

            class TestMyTaxSystemStuff(TaxSystemTestCase):
                def test_should_do_what_i_need(self):
                    user = self.user
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Initialize Alliance Auth test data
        load_allianceauth()

        # Request Factory
        cls.factory = RequestFactory()

        # User with Standard Access - Corporation 2001
        cls.user, cls.user_character = create_user_from_evecharacter(
            character_id=1001,
            permissions=["taxsystem.basic_access"],
        )
        # User with Standard Access - Corporation 2002
        cls.user2, cls.user2_character = create_user_from_evecharacter(
            character_id=1002,
            permissions=["taxsystem.basic_access"],
        )
        # User with Superuser Access - Corporation 2003
        cls.superuser, cls.superuser_character = create_user_from_evecharacter(
            character_id=1003,
            permissions=[],
        )
        cls.superuser.is_superuser = True
        cls.superuser.save()
        # User with Manage Own Corporation Access - Corporation 2001
        cls.manage_own_user, cls.manage_own_character = create_user_from_evecharacter(
            character_id=1004,
            permissions=[
                "taxsystem.basic_access",
                "taxsystem.manage_own_corp",
                "taxsystem.manage_own_alliance",
            ],
        )
        # User with Manage Corporations Access - Corporation 2001
        cls.manage_user, cls.manage_character = create_user_from_evecharacter(
            character_id=1005,
            permissions=[
                "taxsystem.basic_access",
                "taxsystem.manage_corps",
                "taxsystem.manage_alliances",
            ],
        )



# Generated by Django 4.2.26 on 2025-11-20 13:03

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taxsystem", "0015_alter_general_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="allianceupdatestatus",
            name="section",
            field=models.CharField(
                choices=[
                    ("payment_system", "Payment System"),
                    ("payments", "Payments"),
                    ("payday", "Payday"),
                ],
                db_index=True,
                max_length=32,
            ),
        ),
        migrations.AlterField(
            model_name="corporationupdatestatus",
            name="section",
            field=models.CharField(
                choices=[
                    ("division_names", "Wallet Division Names"),
                    ("division", "Wallet Division"),
                    ("wallet", "Wallet Journal"),
                    ("members", "Members"),
                    ("payment_system", "Payment System"),
                    ("payments", "Payments"),
                    ("payday", "Payday"),
                ],
                db_index=True,
                max_length=32,
            ),
        ),
        migrations.RemoveField(
            model_name="corporationpayments",
            name="corporation_id",
        ),
        migrations.AddField(
            model_name="corporationpayments",
            name="owner_id",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="ID of the owner (corporation or alliance) associated with this payment",
                null=True,
            ),
        ),
        migrations.AddIndex(
            model_name="alliancepayments",
            index=models.Index(
                fields=["account", "owner_id", "request_status", "-date"],
                name="taxsystem_a_account_677604_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="alliancepayments",
            index=models.Index(
                fields=["request_status", "-date"],
                name="taxsystem_a_request_d243c8_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="corporationpayments",
            index=models.Index(
                fields=["account", "owner_id", "request_status", "-date"],
                name="taxsystem_c_account_dafabb_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="corporationpayments",
            index=models.Index(
                fields=["request_status", "-date"],
                name="taxsystem_c_request_5fb377_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="members",
            index=models.Index(
                fields=["owner", "character_name"],
                name="taxsystem_m_owner_i_f6738f_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="members",
            index=models.Index(fields=["status"], name="taxsystem_m_status_38e108_idx"),
        ),
    ]



"""Forms for the taxsystem app."""

# Django
from django import forms
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# AA TaxSystem
from taxsystem.models.alliance import (
    AllianceAdminHistory,
    AlliancePaymentHistory,
)
from taxsystem.models.corporation import (
    CorporationAdminHistory,
    CorporationFilter,
    CorporationPaymentHistory,
)
from taxsystem.models.helpers.textchoices import FilterMatchType


def get_mandatory_form_label_text(text: str) -> str:
    """Label text for mandatory form fields"""

    required_marker = "<span class='form-required-marker'>*</span>"

    return mark_safe(
        f"<span class='form-field-required'>{text} {required_marker}</span>"
    )


class AcceptCorporationPaymentForm(forms.ModelForm):
    """Form for accepting corporation payment."""

    class Meta:
        model = CorporationPaymentHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for accepting the payment"),
        }
        labels = {
            "comment": _("Comment (optional)"),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                }
            ),
        }


class AcceptAlliancePaymentForm(forms.ModelForm):
    """Form for accepting alliance payment."""

    class Meta:
        model = AlliancePaymentHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for accepting the payment"),
        }
        labels = {
            "comment": _("Comment (optional)"),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                }
            ),
        }


class RejectCorporationPaymentForm(forms.ModelForm):
    """Form for corporation payment rejecting."""

    class Meta:
        model = CorporationPaymentHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for rejecting the payment"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Reject Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class RejectAlliancePaymentForm(forms.ModelForm):
    """Form for alliance payment rejecting."""

    class Meta:
        model = AlliancePaymentHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for rejecting the payment"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Reject Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class UndoCorporationPaymentForm(forms.ModelForm):
    """Form for corporation payment undoing."""

    class Meta:
        model = CorporationPaymentHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for undoing the payment"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Undo Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class UndoAlliancePaymentForm(forms.ModelForm):
    """Form for alliance payment undoing."""

    class Meta:
        model = AlliancePaymentHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for undoing the payment"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Undo Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class DeleteCorporationPaymentForm(forms.ModelForm):
    """Form for corporation payment deleting."""

    class Meta:
        model = CorporationPaymentHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for deleting the payment"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Delete Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class DeleteAlliancePaymentForm(forms.ModelForm):
    """Form for alliance payment deleting."""

    class Meta:
        model = AlliancePaymentHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for deleting the payment"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Delete Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class PaymentAddForm(forms.Form):
    """Form for payment adding."""

    amount = forms.IntegerField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Amount")),
        help_text=_("Amount to be added"),
        widget=forms.NumberInput(),
    )

    comment = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Add Reason")),
        help_text=_("Reason for adding this payment"),
        widget=forms.Textarea(attrs={"rows": 5}),
    )


class DeleteMemberForm(forms.ModelForm):
    """Form for member deleting."""

    class Meta:
        model = CorporationAdminHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for deleting the member"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Delete Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class DeleteCorporationFilterForm(forms.ModelForm):
    """Form for deleting Filter."""

    class Meta:
        model = CorporationAdminHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for deleting the filter"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Delete Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class DeleteCorporationFilterSetForm(forms.ModelForm):
    """Form for deleting Filter Set."""

    class Meta:
        model = CorporationAdminHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for deleting the filter set"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Delete Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class DeleteAllianceFilterForm(forms.ModelForm):
    """Form for deleting Filter."""

    class Meta:
        model = AllianceAdminHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for deleting the filter"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Delete Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class DeleteAllianceFilterSetForm(forms.ModelForm):
    """Form for deleting Filter Set."""

    class Meta:
        model = AllianceAdminHistory
        fields = ["comment"]
        help_texts = {
            "comment": _("Reason for deleting the filter set"),
        }
        labels = {
            "comment": get_mandatory_form_label_text(text=_("Delete Reason")),
        }
        widgets = {
            "comment": forms.Textarea(
                attrs={
                    "rows": 5,
                    "required": "required",
                }
            ),
        }


class AddJournalFilterForm(forms.Form):
    filter_set = forms.ModelChoiceField(
        queryset=None,
        label=_("Filter Set"),
        required=True,
    )
    filter_type = forms.ChoiceField(
        choices=CorporationFilter.FilterType.choices,
        label=_("Filter Type"),
        required=True,
    )
    match_type = forms.ChoiceField(
        choices=FilterMatchType.choices,
        label=_("Match Type"),
        required=True,
    )
    value = forms.CharField(
        label=_("Filter Value"),
        required=True,
        widget=forms.TextInput(attrs={"placeholder": _("Enter filter value")}),
    )

    def __init__(self, *args, queryset=None, **kwargs):
        super().__init__(*args, **kwargs)
        if queryset is not None:
            self.fields["filter_set"].queryset = queryset


class CreateFilterSetForm(forms.Form):
    name = forms.CharField(
        label=_("Filter Set Name"),
        required=True,
        widget=forms.TextInput(attrs={"placeholder": _("Enter filter set name")}),
    )
    description = forms.CharField(
        label=_("Filter Set Description"),
        required=False,
        widget=forms.Textarea(
            attrs={"placeholder": _("Enter filter set description"), "rows": 3}
        ),
    )



# Generated by Django 4.2.26 on 2026-01-11 12:42

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taxsystem", "0023_alliancefilter_match_type_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="allianceadminhistory",
            name="action",
            field=models.CharField(
                choices=[
                    ("", ""),
                    ("Added", "Added"),
                    ("Changed", "Changed"),
                    ("Deleted", "Deleted"),
                ],
                default="",
                help_text="Action performed",
                max_length=20,
                verbose_name="Action",
            ),
        ),
        migrations.AlterField(
            model_name="alliancepaymentaccount",
            name="status",
            field=models.CharField(
                blank=True,
                choices=[
                    ("active", "Active"),
                    ("inactive", "Inactive"),
                    ("deactivated", "Deactivated"),
                    ("missing", "Missing"),
                ],
                default="active",
                max_length=16,
            ),
        ),
        migrations.AlterField(
            model_name="alliancepaymenthistory",
            name="new_status",
            field=models.CharField(
                choices=[
                    ("approved", "Approved"),
                    ("pending", "Pending"),
                    ("rejected", "Rejected"),
                    ("needs_approval", "Requires Auditor"),
                ],
                help_text="New Status of the action",
                max_length=20,
                verbose_name="New Status",
            ),
        ),
        migrations.AlterField(
            model_name="alliancepayments",
            name="request_status",
            field=models.CharField(
                choices=[
                    ("approved", "Approved"),
                    ("pending", "Pending"),
                    ("rejected", "Rejected"),
                    ("needs_approval", "Requires Auditor"),
                ],
                default="pending",
                max_length=16,
                verbose_name="Request Status",
            ),
        ),
        migrations.AlterField(
            model_name="corporationadminhistory",
            name="action",
            field=models.CharField(
                choices=[
                    ("", ""),
                    ("Added", "Added"),
                    ("Changed", "Changed"),
                    ("Deleted", "Deleted"),
                ],
                default="",
                help_text="Action performed",
                max_length=20,
                verbose_name="Action",
            ),
        ),
        migrations.AlterField(
            model_name="corporationpaymentaccount",
            name="status",
            field=models.CharField(
                blank=True,
                choices=[
                    ("active", "Active"),
                    ("inactive", "Inactive"),
                    ("deactivated", "Deactivated"),
                    ("missing", "Missing"),
                ],
                default="active",
                max_length=16,
            ),
        ),
        migrations.AlterField(
            model_name="corporationpaymenthistory",
            name="new_status",
            field=models.CharField(
                choices=[
                    ("approved", "Approved"),
                    ("pending", "Pending"),
                    ("rejected", "Rejected"),
                    ("needs_approval", "Requires Auditor"),
                ],
                help_text="New Status of the action",
                max_length=20,
                verbose_name="New Status",
            ),
        ),
        migrations.AlterField(
            model_name="corporationpayments",
            name="request_status",
            field=models.CharField(
                choices=[
                    ("approved", "Approved"),
                    ("pending", "Pending"),
                    ("rejected", "Rejected"),
                    ("needs_approval", "Requires Auditor"),
                ],
                default="pending",
                max_length=16,
                verbose_name="Request Status",
            ),
        ),
        migrations.AlterField(
            model_name="members",
            name="status",
            field=models.CharField(
                blank=True,
                choices=[
                    ("active", "Active"),
                    ("missing", "Missing"),
                    ("noaccount", "Unregistered"),
                    ("is_alt", "is alt"),
                ],
                default="active",
                max_length=10,
                verbose_name="Status",
            ),
        ),
    ]



"""This module provides lazy loading of some common functions and objects that are not needed for every request."""

# Django
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import (
    alliance_logo_url,
    character_portrait_url,
    corporation_logo_url,
    type_render_url,
)


def get_character_portrait_url(
    character_id: int, size: int = 32, character_name: str = None, as_html: bool = False
) -> str:
    """Get the character portrait for a character ID."""
    try:
        render_url = character_portrait_url(character_id=character_id, size=size)
    except ValueError:
        return ""

    if as_html:
        render_html = format_html(
            '<img class="character-portrait rounded-circle" src="{}" alt="{}">',
            render_url,
            character_name,
        )
        return render_html
    return render_url


def get_corporation_logo_url(
    corporation_id: int,
    size: int = 32,
    corporation_name: str = None,
    as_html: bool = False,
) -> str:
    """Get the corporation logo for a corporation ID."""

    render_url = corporation_logo_url(corporation_id=corporation_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="corporation-logo rounded-circle" src="{}" alt="{}">',
            render_url,
            corporation_name,
        )
        return render_html
    return render_url


def get_alliance_logo_url(
    alliance_id: int,
    size: int = 32,
    alliance_name: str = None,
    as_html: bool = False,
) -> str:
    """Get the alliance logo for a alliance ID."""

    render_url = alliance_logo_url(alliance_id=alliance_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="alliance-logo rounded-circle" src="{}" alt="{}">',
            render_url,
            alliance_name,
        )
        return render_html
    return render_url


def get_type_render_url(
    type_id: int, size: int = 32, type_name: str = None, as_html: bool = False
) -> str:
    """Get the type render for a type ID."""

    render_url = type_render_url(type_id=type_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="type-render rounded-circle" src="{}" alt="{}">',
            render_url,
            type_name,
        )
        return render_html
    return render_url



"""
Example tests demonstrating usage of the ESI OpenAPI stub system.
"""

# Standard Library
from unittest.mock import patch

# Alliance Auth
from esi.exceptions import HTTPClientError, HTTPNotModified, HTTPServerError

# AA TaxSystem
# AA Taxsystem
from taxsystem.tests import NoSocketsTestCase
from taxsystem.tests.testdata.esi_stub_openapi import (
    EsiEndpoint,
    create_esi_client_stub,
)

# Example test data configuration
EXAMPLE_TEST_DATA = {
    "Skills": {
        "GetCharactersCharacterIdSkills": {
            "skills": [
                {"skill_id": 12345, "trained_skill_level": 5, "active_skill_level": 5},
                {"skill_id": 67890, "trained_skill_level": 3, "active_skill_level": 3},
            ],
            "total_sp": 50000000,
            "unallocated_sp": 100000,
        },
        "GetCharactersCharacterIdSkillqueue": [
            {
                "skill_id": 11111,
                "finished_level": 4,
                "queue_position": 0,
                "finish_date": "2025-12-31T23:59:59Z",
                "start_date": "2025-12-01T00:00:00Z",
                "training_start_sp": 100000,
                "level_start_sp": 50000,
                "level_end_sp": 150000,
            },
            {
                "skill_id": 22222,
                "finished_level": 5,
                "queue_position": 1,
                "finish_date": "2026-01-15T12:00:00Z",
                "start_date": "2025-12-31T23:59:59Z",
                "training_start_sp": 200000,
                "level_start_sp": 150000,
                "level_end_sp": 300000,
            },
        ],
    },
    "Character": {
        "GetCharactersCharacterId": {
            "character_id": 12345678,
            "name": "Test Character",
            "corporation_id": 98765432,
            "birthday": "2015-03-24T11:37:00Z",
        },
    },
}

# Example endpoints for basic tests
EXAMPLE_ENDPOINTS = [
    EsiEndpoint("Character", "GetCharactersCharacterId", "character_id"),
    EsiEndpoint("Skills", "GetCharactersCharacterIdSkills", "character_id"),
    EsiEndpoint("Skills", "GetCharactersCharacterIdSkillqueue", "character_id"),
]


class TestEsiStubUsage(NoSocketsTestCase):
    """Example tests showing how to use the ESI stub system."""

    def test_should_return_single_result(self):
        """
        Test should return single result using result() method.
        """
        # Create a stub with example data and endpoints
        stub = create_esi_client_stub(EXAMPLE_TEST_DATA, endpoints=EXAMPLE_ENDPOINTS)
        # Simulate an ESI call that returns a single result
        operation = stub.Character.GetCharactersCharacterId(character_id=12345678)
        result = operation.result()
        # Verify the data - now using attributes instead of dict keys
        self.assertEqual(result.character_id, 12345678)
        self.assertEqual(result.name, "Test Character")
        self.assertEqual(result.corporation_id, 98765432)

    def test_should_return_list(self):
        """
        Test should return list of results using results() method.
        """
        # Create a stub with example data and endpoints
        stub = create_esi_client_stub(EXAMPLE_TEST_DATA, endpoints=EXAMPLE_ENDPOINTS)
        # Simulate an ESI call that returns a list of results
        operation = stub.Skills.GetCharactersCharacterIdSkillqueue(
            character_id=12345678
        )
        results = operation.results()
        # Verify the data - now using attributes
        self.assertIsInstance(results, list)
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0].skill_id, 11111)
        self.assertEqual(results[1].skill_id, 22222)

    def test_should_create_stub_with_custom_data(self):
        """
        Test should create stub with custom test data and return custom values.
        """
        # Define custom test data
        custom_data = {
            "Skills": {
                "GetCharactersCharacterIdSkills": {
                    "skills": [
                        {
                            "skill_id": 99999,
                            "trained_skill_level": 5,
                            "active_skill_level": 5,
                        }
                    ],
                    "total_sp": 1000000,
                    "unallocated_sp": 0,
                }
            }
        }
        # Define endpoints
        endpoints = [
            EsiEndpoint("Skills", "GetCharactersCharacterIdSkills", "character_id"),
        ]
        # Create stub with custom data and endpoints
        stub = create_esi_client_stub(custom_data, endpoints=endpoints)
        # Use the stub
        operation = stub.Skills.GetCharactersCharacterIdSkills(character_id=12345)
        result = operation.result()
        # Verify custom data is returned - using attributes
        self.assertEqual(result.total_sp, 1000000)
        self.assertEqual(len(result.skills), 1)
        self.assertEqual(result.skills[0].skill_id, 99999)

    @patch("taxsystem.providers.esi")
    def test_should_mock_esi(self, mock_esi):
        """
        Test should mock esi provider to return the stub client.
        """
        # Create a stub with endpoints
        stub = create_esi_client_stub(EXAMPLE_TEST_DATA, endpoints=EXAMPLE_ENDPOINTS)
        # Make mock.client return our stub
        type(mock_esi).client = property(lambda self: stub)
        # Now when code calls esi.client, it will get our stub - Verify it works
        self.assertEqual(mock_esi.client, stub)

    def test_should_handle_nested_data_structures(self):
        """
        Test should handle stub with nested data structures correctly.
        """
        # Create stub with example data and endpoints
        stub = create_esi_client_stub(EXAMPLE_TEST_DATA, endpoints=EXAMPLE_ENDPOINTS)
        # Simulate an ESI call that returns nested data
        operation = stub.Skills.GetCharactersCharacterIdSkills(character_id=12345)
        result = operation.result()
        # Verify nested structure - using attributes
        self.assertTrue(hasattr(result, "skills"))
        self.assertIsInstance(result.skills, list)
        self.assertEqual(result.skills[0].skill_id, 12345)

    def test_should_support_dynamic_callable_data(self):
        """
        Test should support callable/dynamic test data based on input parameters.
        """

        def dynamic_skill_data(**kwargs):
            """Return dynamic data based on skill_points parameter."""
            skill_points = kwargs.get("skill_points", 0)
            return {
                "total_sp": skill_points,  # SP based on input parameter
                "skills": [],
                "unallocated_sp": skill_points // 10,  # 10% unallocated
            }

        # Create stub with callable data - using a custom parameter
        custom_data = {
            "Skills": {
                "GetCharactersCharacterIdSkills": dynamic_skill_data,
            }
        }
        endpoints = [
            EsiEndpoint("Skills", "GetCharactersCharacterIdSkills", "skill_points"),
        ]
        stub = create_esi_client_stub(custom_data, endpoints=endpoints)
        # Call with different skill_points values
        op1 = stub.Skills.GetCharactersCharacterIdSkills(skill_points=1000000)
        result1 = op1.result()
        op2 = stub.Skills.GetCharactersCharacterIdSkills(skill_points=2000000)
        result2 = op2.result()
        # Verify dynamic data works - using attributes
        self.assertEqual(result1.total_sp, 1000000)
        self.assertEqual(result1.unallocated_sp, 100000)
        self.assertEqual(result2.total_sp, 2000000)
        self.assertEqual(result2.unallocated_sp, 200000)

    def test_should_raise_attribute_error_for_unregistered_method(self):
        """
        Test should raise AttributeError when calling unregistered method.
        """
        endpoints = [
            EsiEndpoint("Skills", "GetCharactersCharacterIdSkills", "character_id"),
        ]
        stub = create_esi_client_stub({"Skills": {}}, endpoints=endpoints)
        # Call a method that isn't registered should raise AttributeError
        with self.assertRaises(AttributeError) as context:
            stub.Skills.SomeUnconfiguredMethod(character_id=12345)
        self.assertIn("not registered", str(context.exception))

    def test_should_wrap_single_item_in_list(self):
        """
        Test should wrap single items in a list when using results() method.
        """
        custom_data = {
            "Test": {
                "SingleItemMethod": {"id": 1, "name": "single"},
            }
        }
        endpoints = [
            EsiEndpoint("Test", "SingleItemMethod", "id"),
        ]
        stub = create_esi_client_stub(custom_data, endpoints=endpoints)

        operation = stub.Test.SingleItemMethod()
        results = operation.results()

        # Single item should be wrapped in list - using attributes
        self.assertIsInstance(results, list)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, 1)

    def test_should_return_tuple_with_response_on_result(self):
        """
        Test should return tuple with data and response when return_response=True for result() method.
        """

        stub = create_esi_client_stub(EXAMPLE_TEST_DATA, endpoints=EXAMPLE_ENDPOINTS)

        operation = stub.Character.GetCharactersCharacterId(character_id=12345678)
        data, response = operation.result(return_response=True)

        self.assertEqual(data.character_id, 12345678)
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.headers, dict)

    def test_should_return_tuple_with_response_on_results(self):
        """
        Test should return tuple with list data and response for results() method.
        """

        stub = create_esi_client_stub(EXAMPLE_TEST_DATA, endpoints=EXAMPLE_ENDPOINTS)

        operation = stub.Skills.GetCharactersCharacterIdSkillqueue(
            character_id=12345678
        )
        data, response = operation.results(return_response=True)

        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 2)
        self.assertEqual(response.status_code, 200)

    def test_should_raise_http_not_modified_exception(self):
        """
        Test should raise HTTPNotModified exception via side_effect parameter.
        """
        # Define endpoints with side effects
        _endpoints = [
            EsiEndpoint(
                "Character",
                "GetCharactersCharacterId",
                "character_id",
                side_effect=HTTPNotModified(304, {}),
            ),
        ]
        test_data = {
            "Character": {
                "GetCharactersCharacterId": {"character_id": 12345, "name": "Test"}
            }
        }
        stub = create_esi_client_stub(test_data, endpoints=_endpoints)
        operation = stub.Character.GetCharactersCharacterId(character_id=12345)
        with self.assertRaises(HTTPNotModified):
            operation.result()

    def test_should_raise_http_client_error_exception(self):
        """
        Test should raise HTTPClientError exception via side_effect parameter.
        """
        _endpoints = [
            EsiEndpoint(
                "Skills",
                "GetCharactersCharacterIdSkills",
                "character_id",
                side_effect=HTTPClientError(404, {}, b"Not Found"),
            ),
        ]
        test_data = {
            "Skills": {"GetCharactersCharacterIdSkills": {"total_sp": 0, "skills": []}}
        }
        stub = create_esi_client_stub(test_data, endpoints=_endpoints)
        operation = stub.Skills.GetCharactersCharacterIdSkills(character_id=12345)
        with self.assertRaises(HTTPClientError):
            operation.result()

    def test_should_raise_http_server_error_exception(self):
        """
        Test should raise HTTPServerError exception via side_effect parameter.
        """
        _endpoints = [
            EsiEndpoint(
                "Skills",
                "GetCharactersCharacterIdSkillqueue",
                "character_id",
                side_effect=HTTPServerError(500, {}, b"Server Error"),
            ),
        ]
        test_data = {"Skills": {"GetCharactersCharacterIdSkillqueue": []}}
        stub = create_esi_client_stub(test_data, endpoints=_endpoints)
        operation = stub.Skills.GetCharactersCharacterIdSkillqueue(character_id=12345)
        with self.assertRaises(HTTPServerError):
            operation.results()

    def test_side_effect_os_error(self):
        """
        Test that OSError exception can be simulated via endpoints.
        """
        _endpoints = [
            EsiEndpoint(
                "Character",
                "GetCharactersCharacterId",
                "character_id",
                side_effect=OSError("Connection timeout"),
            ),
        ]

        test_data = {"Character": {"GetCharactersCharacterId": {"character_id": 12345}}}

        stub = create_esi_client_stub(test_data, endpoints=_endpoints)
        operation = stub.Character.GetCharactersCharacterId(character_id=12345)

        # Should raise OSError
        with self.assertRaises(OSError):
            operation.result()

    def test_endpoints_restrict_available_methods(self):
        """
        Test that only registered endpoints are available when endpoints are provided.
        """
        _endpoints = [
            EsiEndpoint(
                "Character",
                "GetCharactersCharacterId",
                "character_id",
            ),
        ]

        test_data = {
            "Character": {
                "GetCharactersCharacterId": {"character_id": 12345, "name": "Test"},
                "GetCharactersCharacterIdRoles": {"roles": []},  # Not registered
            }
        }

        stub = create_esi_client_stub(test_data, endpoints=_endpoints)

        # Registered endpoint should work
        operation = stub.Character.GetCharactersCharacterId(character_id=12345)
        result = operation.result()
        self.assertEqual(result.character_id, 12345)

        # Non-registered endpoint should raise AttributeError
        with self.assertRaises(AttributeError) as context:
            stub.Character.GetCharactersCharacterIdRoles(character_id=12345)

        self.assertIn("not registered", str(context.exception))

    def test_endpoints_restrict_available_categories(self):
        """
        Test that only categories with registered endpoints are available.
        """
        _endpoints = [
            EsiEndpoint(
                "Character",
                "GetCharactersCharacterId",
                "character_id",
            ),
        ]

        test_data = {
            "Character": {
                "GetCharactersCharacterId": {"character_id": 12345, "name": "Test"}
            },
            "Skills": {
                "GetCharactersCharacterIdSkills": {
                    "total_sp": 0
                }  # Category not registered
            },
        }

        stub = create_esi_client_stub(test_data, endpoints=_endpoints)

        # Registered category should work
        operation = stub.Character.GetCharactersCharacterId(character_id=12345)
        result = operation.result()
        self.assertEqual(result.name, "Test")

        # Non-registered category should raise AttributeError
        with self.assertRaises(AttributeError) as context:
            stub.Skills.GetCharactersCharacterIdSkills(character_id=12345)

        self.assertIn("not registered", str(context.exception))

    def test_endpoints_are_required(self):
        """
        Test that endpoints parameter is required.
        """
        test_data = {
            "Character": {
                "GetCharactersCharacterId": {"character_id": 12345, "name": "Test"}
            }
        }

        # Should raise ValueError when no endpoints provided
        with self.assertRaises(ValueError) as context:
            create_esi_client_stub(test_data)

        self.assertIn("endpoints parameter is required", str(context.exception))

    def test_multiple_endpoints_mixed(self):
        """
        Test multiple endpoints with and without side effects.
        """
        _endpoints = [
            EsiEndpoint(
                "Character",
                "GetCharactersCharacterId",
                "character_id",
                side_effect=HTTPNotModified(304, {}),
            ),
            EsiEndpoint(
                "Skills",
                "GetCharactersCharacterIdSkills",
                "character_id",
            ),
        ]

        test_data = {
            "Character": {"GetCharactersCharacterId": {"character_id": 12345}},
            "Skills": {
                "GetCharactersCharacterIdSkills": {"total_sp": 5000000, "skills": []}
            },
        }

        stub = create_esi_client_stub(test_data, endpoints=_endpoints)

        # Character endpoint should raise exception
        char_op = stub.Character.GetCharactersCharacterId(character_id=12345)
        with self.assertRaises(HTTPNotModified):
            char_op.result()

        # Skills endpoint should return normal data
        skills_op = stub.Skills.GetCharactersCharacterIdSkills(character_id=12345)
        skills_result = skills_op.result()
        self.assertEqual(skills_result.total_sp, 5000000)



# Standard Library
import json

# Third Party
from ninja import NinjaAPI, Schema

# Django
from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.handlers.wsgi import WSGIRequest
from django.db import IntegrityError, transaction
from django.shortcuts import get_object_or_404
from django.utils import timezone
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# AA TaxSystem
from taxsystem import __title__, forms
from taxsystem.api.helpers import core
from taxsystem.api.helpers.icons import (
    get_taxsystem_manage_payments_action_icons,
    get_taxsystem_payments_action_icons,
)
from taxsystem.api.schema import (
    CharacterSchema,
    MembersSchema,
    OwnerSchema,
    PaymentHistorySchema,
    PaymentSchema,
    RequestStatusSchema,
)
from taxsystem.helpers import lazy
from taxsystem.models.corporation import (
    CorporationOwner,
)
from taxsystem.models.helpers.textchoices import (
    AccountStatus,
    AdminActions,
    PaymentActions,
    PaymentRequestStatus,
)
from taxsystem.providers import AppLogger

logger = AppLogger(get_extension_logger(__name__), __title__)


class PaymentCorporationSchema(PaymentSchema):
    character: CharacterSchema


class MembersResponse(Schema):
    corporation: list[MembersSchema]


class TaxAccountSchema(Schema):
    account_id: int
    account_name: str
    account_status: str
    character: CharacterSchema
    payment_pool: int


class PaymentsDetailsResponse(Schema):
    owner: OwnerSchema
    account: TaxAccountSchema
    payment: PaymentSchema
    payment_histories: list[PaymentHistorySchema]


class PaymentsApiEndpoints:
    tags = ["Payments"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "owner/{owner_id}/view/payments/",
            response={200: list, 403: dict, 404: dict},
            tags=self.tags,
        )
        def get_payments(request: WSGIRequest, owner_id: int):
            """
            This Endpoint retrieves all payments to the accordicng owner.
            It checks for the owner's existence and the user's permissions
            before fetching and returning the payment data.

            Args:
                request (WSGIRequest): The incoming HTTP request.
                owner_id (int): The ID of the owner whose payments are to be retrieved.
            Returns:
                A list of payment data if successful, or an error message with appropriate status code.
            """
            owner, perms = core.get_owner(request, owner_id)

            if owner is None:
                return 404, {"error": "Owner not Found."}

            if perms is False:
                return 403, {"error": "Permission Denied."}

            # Get Payments
            payments = (
                owner.payment_model.objects.get_visible(user=request.user)
                .select_related(
                    "account",
                    "account__user",
                    "account__user__profile",
                    "account__user__profile__main_character",
                )
                .order_by("-date")
            )

            # TODO for Larger datasets implement pagination and server side DataTables
            # Limit to last 10,000 payments
            payments = payments[:10000]

            response_payments_list: list[PaymentCorporationSchema] = []
            for payment in payments:
                character_portrait = lazy.get_character_portrait_url(
                    payment.character_id, size=32, as_html=True
                )
                # Create the action buttons
                actions_html = str(
                    get_taxsystem_payments_action_icons(
                        request=request, payment=payment, checkbox=True
                    )
                )

                # Create the request status
                response_request_status = RequestStatusSchema(
                    status=payment.get_request_status_display(),
                    color=PaymentRequestStatus(payment.request_status).color(),
                )

                response_payment = PaymentCorporationSchema(
                    payment_id=payment.pk,
                    character=CharacterSchema(
                        character_id=payment.character_id,
                        character_name=payment.account.name,
                        character_portrait=character_portrait,
                    ),
                    amount=payment.amount,
                    date=payment.formatted_payment_date,
                    request_status=response_request_status,
                    division_name=payment.division_name,
                    reviser=payment.reviser,
                    reason=payment.reason,
                    actions=actions_html,
                )
                response_payments_list.append(response_payment)
            return response_payments_list

        @api.get(
            "owner/{owner_id}/view/my-payments/",
            response={200: list, 404: dict},
            tags=self.tags,
        )
        def get_my_payments(request: WSGIRequest, owner_id: int):
            """
            This Endpoint retrieves all payments made by the requesting user
            according to the owner. It checks for the owner's existence
            before fetching and returning the payment data.
            Args:
                request (WSGIRequest): The incoming HTTP request.
                owner_id (int): The ID of the owner whose payments are to be retrieved.
            Returns:
                A list of payment data if successful, or an error message with appropriate status code.
            """
            owner = core.get_owner(request, owner_id)[0]

            if owner is None:
                return 404, {"error": "Owner not Found."}

            # Get Payments
            payments = (
                owner.payment_model.objects.filter(
                    owner=owner,
                    account__user=request.user,
                )
                .select_related(
                    "account",
                    "account__user",
                    "account__user__profile",
                    "account__user__profile__main_character",
                )
                .order_by("-date")
            )
            # Limit to last 10,000 payments
            payments = payments[:10000]

            response_payments_list: list[PaymentCorporationSchema] = []
            for payment in payments:
                character_portrait = lazy.get_character_portrait_url(
                    payment.character_id, size=32, as_html=True
                )

                # Create the request status
                response_request_status = RequestStatusSchema(
                    status=payment.get_request_status_display(),
                    color=PaymentRequestStatus(payment.request_status).color(),
                )

                response_payment = PaymentSchema(
                    payment_id=payment.pk,
                    character=CharacterSchema(
                        character_id=payment.character_id,
                        character_name=payment.account.name,
                        character_portrait=character_portrait,
                    ),
                    amount=payment.amount,
                    date=payment.formatted_payment_date,
                    request_status=response_request_status,
                    division_name=payment.division_name,
                    reviser=payment.reviser,
                    reason=payment.reason,
                )
                response_payments_list.append(response_payment)
            return response_payments_list

        @api.get(
            "owner/{owner_id}/payment/{payment_pk}/view/details/",
            response={200: PaymentsDetailsResponse, 403: dict, 404: dict},
            tags=self.tags,
        )
        # pylint: disable=too-many-locals
        def get_payment_details(request, owner_id: int, payment_pk: int):
            owner, perms = core.get_manage_owner(request, owner_id)

            # pylint: disable=duplicate-code
            if owner is None:
                return 404, {"error": _("Owner not Found.")}

            payment = get_object_or_404(owner.payment_model, pk=payment_pk)
            perms = perms or core.get_character_permissions(
                request, payment.character_id
            )

            # pylint: disable=duplicate-code
            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            response_payment_histories: list[PaymentHistorySchema] = []
            payments_history = owner.payment_history_model.objects.filter(
                payment=payment,
            ).order_by("-date")

            # Create a list for the payment histories
            for log in payments_history:
                response_log = PaymentHistorySchema(
                    log_id=log.pk,
                    reviser=log.user.username if log.user else _("System"),
                    date=log.date.strftime("%Y-%m-%d %H:%M:%S"),
                    action=log.get_action_display(),
                    comment=log.comment,
                    status=log.get_new_status_display(),
                )
                response_payment_histories.append(response_log)

            # Create the tax account
            response_account = TaxAccountSchema(
                account_id=payment.account.pk,
                account_name=payment.account.name,
                account_status=AccountStatus(payment.account.status).html(),
                character=CharacterSchema(
                    character_id=payment.character_id,
                    character_name=payment.account.name,
                    character_portrait=lazy.get_character_portrait_url(
                        payment.character_id, size=32, as_html=True
                    ),
                    corporation_id=payment.account.owner.pk,
                    corporation_name=payment.account.owner.name,
                ),
                payment_pool=payment.account.deposit,
            )

            response_request_status = RequestStatusSchema(
                status=payment.get_request_status_display(),
                html=PaymentRequestStatus(payment.request_status).alert(),
            )

            # Create the payment
            response_payment = PaymentSchema(
                payment_id=payment.pk,
                amount=payment.amount,
                date=payment.formatted_payment_date,
                request_status=response_request_status,
                division_name=payment.division_name,
                reason=payment.reason,
                reviser=payment.reviser,
            )

            response_owner = OwnerSchema(
                owner_id=owner.eve_id,
                owner_name=owner.name,
            )

            payment_details_response = PaymentsDetailsResponse(
                owner=response_owner,
                account=response_account,
                payment=response_payment,
                payment_histories=response_payment_histories,
            )

            return payment_details_response

        @api.get(
            "owner/{owner_id}/character/{character_id}/view/payments/",
            response={200: list[PaymentSchema], 403: dict, 404: dict},
            tags=self.tags,
        )
        def get_member_payments(request, owner_id: int, character_id: int):
            owner, perms = core.get_manage_owner(request, owner_id)

            # pylint: disable=duplicate-code
            if owner is None:
                return 404, {"error": _("Corporation Not Found")}

            # pylint: disable=duplicate-code
            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            # Filter payments by character
            payments = owner.payment_model.objects.filter(
                account__user__profile__main_character__character_id=character_id,
                owner=owner,
            ).order_by("-date")
            # Limit to last 10,000 payments
            payments = payments[:10000]

            response_payments_list: list[PaymentSchema] = []
            for payment in payments:
                # Create the actions
                actions_html = str(
                    get_taxsystem_manage_payments_action_icons(
                        request=request, payment=payment
                    )
                )

                # pylint: disable=duplicate-code
                # Create the request status
                response_request_status = RequestStatusSchema(
                    status=payment.get_request_status_display(),
                    color=PaymentRequestStatus(payment.request_status).color(),
                )

                response_payment = PaymentSchema(
                    payment_id=payment.pk,
                    amount=payment.amount,
                    date=payment.formatted_payment_date,
                    request_status=response_request_status,
                    division_name=payment.division_name,
                    reason=payment.reason,
                    actions=actions_html,
                    reviser=payment.reviser,
                )
                response_payments_list.append(response_payment)

            return response_payments_list

        @api.post(
            "owner/{owner_id}/account/{account_pk}/manage/add-payment/",
            response={200: dict, 400: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def add_payment(request: WSGIRequest, owner_id: int, account_pk: int):
            """
            Handle an Request to Add a custom Payment

            This Endpoint adds a custom payment for a tax account.
            It validates the request, checks permissions, and adds the payment to the according tax account.

            Args:
                request (WSGIRequest): The HTTP request object.
                owner_id (int): The ID of the owner whose filter set is to be retrieved.
                account_pk (int): The ID of the tax account to which the payment will be added.
            Returns:
                dict: A dictionary containing the success status and message.
            """
            owner, perms = core.get_manage_owner(request, owner_id)

            # Check if owner exists
            if owner is None:
                return 404, {"error": _("Owner not Found.")}

            # Check permissions
            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            # Validate the form data
            form = forms.PaymentAddForm(data=json.loads(request.body))
            if not form.is_valid():
                msg = _("Invalid form data.")
                return 400, {"success": False, "message": msg}

            amount = form.cleaned_data["amount"]
            comment = form.cleaned_data["comment"]

            # Begin transaction
            try:
                with transaction.atomic():
                    account = owner.account_model.objects.get(
                        owner=owner, pk=account_pk
                    )
                    payment = owner.payment_model(
                        owner=owner,
                        name=account.user.username,
                        journal=None,  # Manual Entry (use NULL to allow multiple manual payments)
                        amount=amount,
                        account=account,
                        date=timezone.now(),
                        reason="",
                        request_status=PaymentRequestStatus.APPROVED,
                        reviser=request.user.username,
                    )

                    # Create log message
                    msg = format_lazy(
                        _("Custom Payment Added: {comment}"),
                        comment=comment,
                    )

                    payment.save()
                    account.deposit += amount
                    account.save()

                    # Log the Payment Action
                    payment.transaction_log(
                        user=request.user,
                        action=PaymentActions.CUSTOM_PAYMENT,
                        new_status=PaymentRequestStatus.APPROVED,
                        comment=comment,
                    ).save()

                return 200, {"success": True, "message": msg}
            except IntegrityError:
                msg = _("Transaction failed. Please try again.")
                return 400, {"success": False, "message": msg}

        @api.post(
            "owner/{owner_id}/payment/{payment_pk}/manage/approve-payment/",
            response={200: dict, 400: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def approve_payment(request: WSGIRequest, owner_id: int, payment_pk: int):
            """
            Handle an Request to Approve a Payment

            This Endpoint approves a payment from an associated tax account.
            It validates the request, checks permissions, and approves the payment to the according tax account.

            Args:
                request (WSGIRequest): The HTTP request object.
                owner_id (int): The ID of the owner whose filter set is to be retrieved.
                payment_pk (int): The ID of the payment to be approved.
            Returns:
                dict: A dictionary containing the success status and message.
            """
            owner, perms = core.get_manage_owner(request, owner_id)

            # Check if owner exists
            if owner is None:
                return 404, {"error": _("Owner not Found.")}

            # Check permissions
            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            payment_form_instance = (
                forms.AcceptCorporationPaymentForm
                if isinstance(owner, CorporationOwner)
                else forms.AcceptAlliancePaymentForm
            )

            # Validate the form data
            form = payment_form_instance(data=json.loads(request.body))
            if not form.is_valid():
                msg = _("Invalid form data.")
                return 400, {"success": False, "message": msg}

            reason = form.cleaned_data["comment"]

            # Begin transaction
            try:
                with transaction.atomic():
                    payment = owner.payment_model.objects.get(
                        pk=payment_pk,
                    )
                    # Check if payment is pending or needs approval
                    if payment.is_pending or payment.is_needs_approval:
                        # Approve Payment
                        payment.request_status = PaymentRequestStatus.APPROVED
                        payment.reviser = (
                            request.user.profile.main_character.character_name
                        )
                        payment.save()

                        # Update Account Deposit
                        payment.account.deposit += payment.amount
                        payment.account.save()

                        # Log the Payment Action
                        payment.transaction_log(
                            user=request.user,
                            action=PaymentActions.STATUS_CHANGE,
                            comment=reason,
                            new_status=PaymentRequestStatus.APPROVED,
                        ).save()

                        # Create response message
                        msg = format_lazy(
                            _(
                                "Payment ID: {pid} - Amount: {amount} - Name: {name} approved"
                            ),
                            pid=payment.pk,
                            amount=intcomma(payment.amount),
                            name=payment.name,
                        )
                        return 200, {"success": True, "message": msg}
                    msg = _("Payment is not pending or does not need approval.")
                    return 400, {"success": True, "message": msg}
            except IntegrityError:
                msg = _("Transaction failed. Please try again.")
                return 400, {"success": False, "message": msg}

        @api.post(
            "owner/{owner_id}/payment/{payment_pk}/manage/undo-payment/",
            response={200: dict, 400: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def undo_payment(request: WSGIRequest, owner_id: int, payment_pk: int):
            """
            Handle an Request to Undo a Payment

            This Endpoint undoes a payment from an associated tax account.
            It validates the request, checks permissions, and undoes the payment to the according tax account.

            Args:
                request (WSGIRequest): The HTTP request object.
                owner_id (int): The ID of the owner whose filter set is to be retrieved.
                payment_pk (int): The ID of the payment to be approved.
            Returns:
                dict: A dictionary containing the success status and message.
            """
            owner, perms = core.get_manage_owner(request, owner_id)

            # Check if owner exists
            if owner is None:
                return 404, {"error": _("Owner not Found.")}

            # Check permissions
            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            payment_form_instance = (
                forms.AcceptCorporationPaymentForm
                if isinstance(owner, CorporationOwner)
                else forms.AcceptAlliancePaymentForm
            )

            # Validate the form data
            form = payment_form_instance(data=json.loads(request.body))
            if not form.is_valid():
                msg = _("Invalid form data.")
                return 400, {"success": False, "message": msg}

            reason = form.cleaned_data["comment"]

            # Begin transaction
            try:
                with transaction.atomic():
                    payment = owner.payment_model.objects.get(
                        pk=payment_pk,
                    )
                    # Check if payment is approved or needs approval
                    if payment.is_approved or payment.is_rejected:
                        if not payment.is_rejected:
                            # Update Account Deposit
                            payment.account.deposit -= payment.amount
                            payment.account.save()
                        payment.request_status = PaymentRequestStatus.PENDING
                        payment.reviser = ""
                        payment.save()

                        # Log the Payment Action
                        payment.transaction_log(
                            user=request.user,
                            action=PaymentActions.STATUS_CHANGE,
                            comment=reason,
                            new_status=PaymentRequestStatus.PENDING,
                        ).save()

                        # Create response message
                        msg = format_lazy(
                            _(
                                "Payment ID: {pid} - Amount: {amount} - Name: {name} undone"
                            ),
                            pid=payment.pk,
                            amount=intcomma(payment.amount),
                            name=payment.name,
                        )
                        return 200, {"success": True, "message": msg}
                    msg = _("Payment is approved or rejected.")
                    return 400, {"success": True, "message": msg}
            except IntegrityError:
                msg = _("Transaction failed. Please try again.")
                return 400, {"success": False, "message": msg}

        @api.post(
            "owner/{owner_id}/payment/{payment_pk}/manage/delete-payment/",
            response={200: dict, 400: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def delete_payment(request: WSGIRequest, owner_id: int, payment_pk: int):
            """
            Handle an Request to Delete a Payment

            This Endpoint deletes a payment from an associated tax account.
            It validates the request, checks permissions, and deletes the payment to the according tax account.

            Args:
                request (WSGIRequest): The HTTP request object.
                owner_id (int): The ID of the owner whose filter set is to be retrieved.
                payment_pk (int): The ID of the payment to be approved.
            Returns:
                dict: A dictionary containing the success status and message.
            """
            owner, perms = core.get_manage_owner(request, owner_id)

            # Check if owner exists
            if owner is None:
                return 404, {"error": _("Owner not Found.")}

            # Check permissions
            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            payment_form_instance = (
                forms.DeleteCorporationPaymentForm
                if isinstance(owner, CorporationOwner)
                else forms.DeleteAlliancePaymentForm
            )

            # Validate the form data
            form = payment_form_instance(data=json.loads(request.body))
            if not form.is_valid():
                msg = _("Invalid form data.")
                return 400, {"success": False, "message": msg}

            reason = form.cleaned_data["comment"]

            # Begin transaction
            try:
                with transaction.atomic():
                    payment = owner.payment_model.objects.get(
                        pk=payment_pk,
                    )
                    if (
                        payment.journal is not None
                    ):  # Prevent deletion of ESI imported payments
                        msg = _("ESI imported payments cannot be deleted")
                        return 400, {"success": False, "message": msg}

                    # Capture values before deletion
                    pid = payment.pk
                    amount = intcomma(payment.amount)
                    name = payment.name

                    # Refund if approved
                    if payment.is_approved:
                        payment.account.deposit -= payment.amount
                        payment.account.save()

                    # Delete Payment
                    payment.delete()

                    msg = format_lazy(
                        _(
                            "Payment ID: {pid} - Amount: {amount} - Name: {name} deleted - {reason}"
                        ),
                        pid=pid,
                        amount=amount,
                        name=name,
                        reason=reason,
                    )

                    # Log the action in Admin History
                    owner.admin_log_model(
                        owner=owner,
                        user=request.user,
                        action=AdminActions.DELETE,
                        comment=msg,
                    ).save()
                    return 200, {"success": True, "message": msg}
            except IntegrityError:
                msg = _("Transaction failed. Please try again.")
                return 400, {"success": False, "message": msg}

        @api.post(
            "owner/{owner_id}/payment/{payment_pk}/manage/reject-payment/",
            response={200: dict, 400: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def reject_payment(request: WSGIRequest, owner_id: int, payment_pk: int):
            """
            Handle an Request to Reject a Payment

            This Endpoint rejects a payment from an associated tax account.
            It validates the request, checks permissions, and rejects the payment to the according tax account.

            Args:
                request (WSGIRequest): The HTTP request object.
                owner_id (int): The ID of the owner whose filter set is to be retrieved.
                payment_pk (int): The ID of the payment to be approved.
            Returns:
                dict: A dictionary containing the success status and message.
            """
            owner, perms = core.get_manage_owner(request, owner_id)

            # Check if owner exists
            if owner is None:
                return 404, {"error": _("Owner not Found.")}

            # Check permissions
            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            payment_form_instance = (
                forms.DeleteCorporationPaymentForm
                if isinstance(owner, CorporationOwner)
                else forms.DeleteAlliancePaymentForm
            )

            # Validate the form data
            form = payment_form_instance(data=json.loads(request.body))
            if not form.is_valid():
                msg = _("Invalid form data.")
                return 400, {"success": False, "message": msg}

            reason = form.cleaned_data["comment"]

            # Begin transaction
            try:
                with transaction.atomic():
                    payment = owner.payment_model.objects.get(
                        pk=payment_pk,
                    )
                    if payment.is_pending or payment.is_needs_approval:
                        payment.request_status = PaymentRequestStatus.REJECTED
                        payment.reviser = (
                            request.user.profile.main_character.character_name
                        )
                        payment.save()

                        msg = format_lazy(
                            _(
                                "Payment ID: {pid} - Amount: {amount} - Name: {name} rejected - {reason}"
                            ),
                            pid=payment.pk,
                            amount=intcomma(payment.amount),
                            name=payment.name,
                            reason=reason,
                        )

                        # Log Admin Action
                        payment.transaction_log(
                            user=request.user,
                            action=PaymentActions.STATUS_CHANGE,
                            comment=reason,
                            new_status=PaymentRequestStatus.REJECTED,
                        ).save()
                        return 200, {"success": True, "message": msg}
                    msg = _("Payment is not pending or does not need approval.")
                    return 400, {"success": True, "message": msg}
            except IntegrityError:
                msg = _("Transaction failed. Please try again.")
                return 400, {"success": False, "message": msg}

        @api.post(
            "owner/{owner_id}/payment/manage/bulk-actions/",
            response={200: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def perform_bulk_actions_payments(request: WSGIRequest, owner_id: int):
            """
            Handle an Request to Bulk Actions

            This Endpoint performs bulk actions for an associated owner.
            It validates the request, checks permissions, and performs the bulk actions accordingly.

            Args:
                request (WSGIRequest): The HTTP request object.
                owner_id (int): The ID of the owner whose filter set is to be retrieved.
            Returns:
                dict: A dictionary containing the success status and message.
            """
            # pylint: disable=duplicate-code
            owner, perms = core.get_manage_owner(request, owner_id)

            # Check if owner exists
            if owner is None:
                return 404, {"error": _("Owner not Found.")}

            # Check permissions
            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            pks_ids = json.loads(request.body).get("pks", [])
            action = json.loads(request.body).get("action", "")

            if len(pks_ids) == 0:
                msg = _("Please select at least one payment to perform bulk actions.")
                return 400, {"success": False, "message": msg}

            if action == "approve":
                status = PaymentRequestStatus.APPROVED
                payments = owner.payment_model.objects.filter(
                    owner=owner,
                    pk__in=pks_ids,
                    request_status__in=[
                        PaymentRequestStatus.PENDING,
                        PaymentRequestStatus.NEEDS_APPROVAL,
                    ],
                )
                with transaction.atomic():
                    runs = 0
                    for payment in payments:
                        # Approve Payment
                        payment.request_status = PaymentRequestStatus.APPROVED
                        payment.reviser = (
                            request.user.profile.main_character.character_name
                        )
                        payment.save()

                        # Update Account Deposit
                        payment.account.deposit += payment.amount
                        payment.account.save()

                        # Log the Payment Action
                        payment.transaction_log(
                            user=request.user,
                            action=PaymentActions.STATUS_CHANGE,
                            comment=_("Bulk Approved"),
                            new_status=PaymentRequestStatus.APPROVED,
                        ).save()
                        runs += 1
            elif action == "reject":
                status = PaymentRequestStatus.REJECTED
                with transaction.atomic():
                    runs = owner.payment_model.objects.filter(
                        owner=owner,
                        pk__in=pks_ids,
                        request_status__in=[
                            PaymentRequestStatus.PENDING,
                            PaymentRequestStatus.NEEDS_APPROVAL,
                        ],
                    ).update(request_status=PaymentRequestStatus.REJECTED)
            else:
                msg = _("Please select a valid action")
                return 400, {"success": False, "message": msg}

            # Create log message
            msg = format_lazy(
                _("Bulk '{status}' performed for {runs} payments ({pks}) for {owner}"),
                runs=runs,
                status=status,
                owner=owner,
                pks=pks_ids,
            )

            # Log Action in Admin History
            owner.admin_log_model(
                user=request.user,
                owner=owner,
                action=AdminActions.CHANGE,
                comment=msg,
            ).save()

            # Return success response
            return 200, {"success": True, "message": msg}
